#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Meeko
#

from abc import ABC, abstractmethod
from copy import deepcopy
from collections import defaultdict
from dataclasses import asdict, dataclass, field
import json
import sys
import warnings

import numpy as np
import rdkit.Chem
from rdkit import Chem
from rdkit.Chem import rdPartialCharges
from rdkit.Chem import rdMolInterchange

from .utils.jsonutils import rdkit_mol_from_json, tuple_to_string, string_to_tuple
from .utils import rdkitutils
from .utils import utils
from .utils.geomutils import calcDihedral
from .utils.pdbutils import PDBAtomInfo
from .receptor_pdbqt import PDBQTReceptor

try:
    from misctools import StereoIsomorphism
except ImportError as _import_misctools_error:
    _has_misctools = False
else:
    _has_misctools = True


from .utils import rdkitutils

# region DEFAULT VALUES
DEFAULT_PDBINFO = None
DEFAULT_CHARGE = 0.0
DEFAULT_COORD = np.array([0.0, 0.0, 0.0], dtype="float")
DEFAULT_ATOMIC_NUM = None
DEFAULT_ATOM_TYPE = None
DEFAULT_IS_IGNORE = False
DEFAULT_GRAPH = []

DEFAULT_BOND_ROTATABLE = False

DEFAULT_RING_CORNER_FLIP = False
DEFAULT_RING_GRAPH = []
DEFAULT_RING_IS_AROMATIC = False
DEFAULT_RING_CLOSURE_BONDS_REMOVED = []
DEFAULT_RING_CLOSURE_PSEUDOS_BY_ATOM = defaultdict
# endregion


# region Helper Data Organization Classes
class UniqAtomParams:
    """
    A helper class used to keep parameters organized in a particular way that lets them be more usable.

    Attributes
    ----------
    params: list[]
        can be thought of as rows
    param_names: list[]
        can be thought of as columns
    """

    def __init__(self):
        self.params = []  # aka rows
        self.param_names = []  # aka column names

    @classmethod
    def from_dict(cls, dictionary):
        """
        Creates an UniqAtomParams object, populates it with information from the input dictionary, then returns
        the new object.

        Parameters
        ----------
        dictionary: dict()
            A dictionary containing the keys "params" and "param_names", where the value for "params" is parseable as
            rows and the value for "param_names" contains the corresponding column data.

        Returns
        -------
        A populated UniqAtomParams object
        """
        uap = UniqAtomParams()
        uap.params = [row.copy() for row in dictionary["params"]]
        uap.param_names = dictionary["param_names"].copy()
        return uap

    def get_indices_from_atom_params(self, atom_params):
        """
        Retrieves the indices of specific atom parameters in the UniqAtomParams object.

        Parameters
        ----------
        atom_params: dict()
            A dict with keys that correspond to the param names already in the UniqAtomParams object. The values are
            lists that should all be the same size, and

        Returns
        -------
        A list of indices corresponding to the order of parameters in the atom_params value lists that indicates the
        index of that "row" of parameters in UniqAtomParams params.
        """
        nr_items = set([len(values) for key, values in atom_params.items()])
        if len(nr_items) != 1:
            raise RuntimeError(
                f"all lists in atom_params must have same length, got {nr_items}"
            )
        if set(atom_params) != set(self.param_names):
            msg = f"parameter names in atom_params differ from internal ones\n"
            msg += f"  - in atom_params: {set(atom_params)}"
            msg += f"  - internal: {set(self.param_names)}"
            raise RuntimeError(msg)
        nr_items = nr_items.pop()
        param_idxs = []
        for i in range(nr_items):
            row = [atom_params[key][i] for key in self.param_names]
            param_index = None
            for j, existing_row in enumerate(self.params):
                if row == existing_row:
                    param_index = j
                    break
            param_idxs.append(param_index)
        return param_idxs

    def add_parameter(self, new_param_dict):
        # remove None values to avoid a column with only Nones
        new_param_dict = {k: v for k, v in new_param_dict.items() if v is not None}
        incoming_keys = set(new_param_dict.keys())
        existing_keys = set(self.param_names)
        new_keys = incoming_keys.difference(existing_keys)
        for new_key in new_keys:
            self.param_names.append(new_key)
            for row in self.params:
                row.append(None)  # fill in empty "cell" in new "column"

        new_row = []
        for key in self.param_names:
            value = new_param_dict.get(key, None)
            new_row.append(value)

        if len(new_keys) == 0:  # try to match with existing row
            for index, row in enumerate(self.params):
                if row == new_row:
                    return index

        # if we are here, we didn't match
        new_row_index = len(self.params)
        self.params.append(new_row)
        return new_row_index

    def add_molsetup(
        self, molsetup, atom_params=None, add_atomic_nr=False, add_atom_type=False
    ):
        if "charge" in molsetup.atom_params or "atom_type" in molsetup.atom_params:
            msg = '"charge" and "atom_type" found in molsetup.atom_params'
            msg += " but are hard-coded to store molsetup.charge and"
            msg += " molsetup.atom_type in the internal data structure"
            raise RuntimeError(msg)
        if atom_params is None:
            atom_params = molsetup.atom_params
        param_idxs = []
        for atom in molsetup.atoms:
            if atom.is_ignore:
                param_idx = None
            else:
                p = {k: v[atom.index] for (k, v) in molsetup.atom_params.items()}
                if add_atomic_nr:
                    if "atomic_nr" in p:
                        raise RuntimeError(
                            "trying to add atomic_nr but it's already in atom_params"
                        )
                    p["atomic_nr"] = atom.atomic_num
                if add_atom_type:
                    if "atom_type" in p:
                        raise RuntimeError(
                            "trying to add atom_type but it's already in atom_params"
                        )
                    p["atom_type"] = atom.atom_type
                param_idx = self.add_parameter(p)
            param_idxs.append(param_idx)
        return param_idxs


@dataclass
class Atom:
    index: int
    pdbinfo: str or PDBAtomInfo = DEFAULT_PDBINFO
    charge: float = DEFAULT_CHARGE
    coord: np.ndarray = field(default_factory=np.ndarray)
    atomic_num: int = DEFAULT_ATOMIC_NUM
    atom_type: str = DEFAULT_ATOM_TYPE
    is_ignore: bool = DEFAULT_IS_IGNORE
    graph: list[int] = field(default_factory=list)
    interaction_vectors: list[np.array] = field(default_factory=list)

    is_dummy: bool = False
    is_pseudo_atom: bool = False

    @staticmethod
    def from_json(obj: dict):
        """
        Takes an object and attempts to deserialize it into an Atom object.

        Parameters
        ----------
        obj: Object
            This can be any object, but it should be a dictionary constructed by deserializing the JSON representation
            of an Atom object.

        Returns
        -------
        If the input is a dictionary corresponding to an Atom, will return an Atom with data populated from the
        dictionary. Otherwise, returns the input object.
        """
        # if the input object is not a dict, we know that it will not be parsable and is unlikely to be usable or
        # safe data, so we should ignore it.
        if type(obj) is not dict:
            return obj

        # Check that all the keys we expect are in the object dictionary as a safety measure
        expected_json_keys = {
            "index",
            "pdbinfo",
            "charge",
            "coord",
            "atomic_num",
            "atom_type",
            "is_ignore",
            "graph",
            "interaction_vectors",
            "is_dummy",
            "is_pseudo_atom",
        }
        if set(obj.keys()) != expected_json_keys:
            return obj

        # Constructs an atom object from the provided keys.
        index = obj["index"]
        pdbinfo = PDBAtomInfo(*obj["pdbinfo"])
        charge = obj["charge"]
        coord = np.asarray(obj["coord"])
        atomic_num = obj["atomic_num"]
        atom_type = obj["atom_type"]
        is_ignore = obj["is_ignore"]
        graph = obj["graph"]
        interaction_vectors = [np.asarray(i) for i in obj["interaction_vectors"]]
        is_dummy = obj["is_dummy"]
        is_pseudo_atom = obj["is_pseudo_atom"]
        output_atom = Atom(
            index,
            pdbinfo,
            charge,
            coord,
            atomic_num,
            atom_type,
            is_ignore,
            graph,
            interaction_vectors,
            is_dummy,
            is_pseudo_atom,
        )
        return output_atom


@dataclass
class Bond:
    canon_id: (int, int)
    index1: int
    index2: int
    rotatable: bool = DEFAULT_BOND_ROTATABLE

    def __init__(
        self,
        index1: int,
        index2: int,
        rotatable: bool = DEFAULT_BOND_ROTATABLE,
    ):
        self.canon_id = self.get_bond_id(index1, index2)
        self.index1 = index1
        self.index2 = index2
        self.rotatable = rotatable
        return

    @staticmethod
    def get_bond_id(idx1: int, idx2: int):
        """
        Generates a consistent, "canonical", bond id from a pair of atom indices in the graph.

        Parameters
        ----------
        idx1: int
            atom index of one of the atoms in the bond
        idx2: int
            atom index of the other atom in the bond

        Returns
        -------
        canon_id: tuple
            a tuple of the two indices in their canonical order.
        """
        idx_min = min(idx1, idx2)
        idx_max = max(idx1, idx2)
        return idx_min, idx_max

    @staticmethod
    def from_json(obj: dict):
        """
        Takes an object and attempts to deserialize it into a Bond object.

        Parameters
        ----------
        obj: Object
            This can be any object, but it should be a dictionary constructed by deserializing the JSON representation
            of a Bond object.

        Returns
        -------
        If the input is a dictionary corresponding to a Bond, will return a Bond with data populated from the
        dictionary. Otherwise, returns the input object.
        """
        # if the input object is not a dict, we know that it will not be parsable and is unlikely to be usable or
        # safe data, so we should ignore it.
        if type(obj) is not dict:
            return obj

        # Check that all the keys we expect are in the object dictionary as a safety measure
        expected_json_keys = {"canon_id", "index1", "index2", "rotatable"}
        if set(obj.keys()) != expected_json_keys:
            return obj

        # Constructs a bond object from the provided keys.
        index1 = obj["index1"]
        index2 = obj["index2"]
        rotatable = obj["rotatable"]
        output_bond = Bond(index1, index2, rotatable)
        return output_bond


@dataclass
class Ring:
    ring_id: tuple
    corner_flip: bool = DEFAULT_RING_CORNER_FLIP
    graph: dict = field(default_factory=list)
    is_aromatic: bool = DEFAULT_RING_IS_AROMATIC

    @staticmethod
    def from_json(obj: dict):
        """
        Takes an object and attempts to deserialize it into a Ring object.

        Parameters
        ----------
        obj: Object
            This can be any object, but it should be a dictionary constructed by deserializing the JSON representation
            of a Ring object.

        Returns
        -------
        If the input is a dictionary corresponding to a Ring, will return a Ring with data populated from the
        dictionary. Otherwise, returns the input object.
        """
        # if the input object is not a dict, we know that it will not be parsable and is unlikely to be usable or
        # safe data, so we should ignore it.
        if type(obj) is not dict:
            return obj

        # Check that all the keys we expect are in the object dictionary as a safety measure
        expected_json_keys = {"ring_id", "corner_flip", "graph", "is_aromatic"}
        if set(obj.keys()) != expected_json_keys:
            return obj

        # Constructs a Ring object from the provided keys.
        ring_id = string_to_tuple(obj["ring_id"], int)
        corner_flip = obj["corner_flip"]
        graph = obj["graph"]
        is_aromatic = obj["is_aromatic"]
        output_ring = Ring(ring_id, corner_flip, graph, is_aromatic)
        return output_ring


@dataclass
class RingClosureInfo:
    bonds_removed: list = field(default_factory=list)
    pseudos_by_atom: dict = DEFAULT_RING_CLOSURE_PSEUDOS_BY_ATOM


@dataclass
class Restraint:
    atom_index: int
    target_coords: (float, float, float)
    kcal_per_angstrom_square: float
    delay_angstroms: float

    def copy(self):
        new_target_coords = (
            self.target_coords[0],
            self.target_coords[1],
            self.target_coords[2],
        )
        new_restraint = Restraint(
            self.atom_index,
            new_target_coords,
            self.kcal_per_angstrom_square,
            self.delay_angstroms,
        )
        return new_restraint

    @staticmethod
    def from_json(obj: dict):
        """
        Takes an object and attempts to deserialize it into a Restraint object.

        Parameters
        ----------
        obj: Object
            This can be any object, but it should be a dictionary constructed by deserializing the JSON representation
            of a Restraint object.

        Returns
        -------
        If the input is a dictionary corresponding to a Restraint, will return a Restraint with data populated from the
        dictionary. Otherwise, returns the input object.
        """
        # if the input object is not a dict, we know that it will not be parsable and is unlikely to be usable or
        # safe data, so we should ignore it.
        if type(obj) is not dict:
            return obj

        # Check that all the keys we expect are in the object dictionary as a safety measure
        expected_json_keys = {
            "atom_index",
            "target_coords",
            "kcal_per_angstrom_square",
            "delay_angstroms",
        }
        if set(obj.keys()) != expected_json_keys:
            return obj

        # Constructs a Restraint object from the provided keys.
        atom_index = obj["atom_index"]
        target_coords = tuple(obj["target_coords"])
        kcal_per_angstrom_square = obj["kcal_per_angstrom_square"]
        delay_angstroms = obj["delay_angstroms"]
        output_restraint = Restraint(
            atom_index, target_coords, kcal_per_angstrom_square, delay_angstroms
        )
        return output_restraint


# endregion


class MoleculeSetup:
    """
    Base MoleculeSetup Class, provides a way to store information about molecules for a number of purposes.

    Attributes
    ----------
    name: str
    is_sidechain: bool
    pseudoatom_count: int

    atoms: list[Atom]
    bond_info: dict[tuple, Bond]
    rings: dict
    ring_closure_info: RingClosureInfo
    rotamers: list[dict]

    atom_params: dict
    restraints: list[Restraint]
    flexibility_model: dict
    """

    # region CLASS CONSTANTS
    PSEUDOATOM_ATOMIC_NUM = 0
    # endregion

    def __init__(self, name: str = None, is_sidechain: bool = False):
        # Molecule Setup Identity
        self.name: str = name
        self.is_sidechain: bool = is_sidechain
        self.pseudoatom_count: int = 0

        # Tracking atoms and bonds
        self.atoms: list[Atom] = []
        self.bond_info: dict[tuple, Bond] = {}
        self.rings: dict[tuple, Ring] = {}
        self.ring_closure_info = RingClosureInfo([], {})
        self.rotamers: list[dict] = []  # TODO: revisit rotamer implementation

        self.atom_params: dict = {}
        self.restraints: list = (
            []
        )  # TODO: determine whether restraints are being used anymore

        # TODO: redesign flexibility model to resolve some of the circular imports and to make it more structured
        self.flexibility_model = None  # from flexibility_model - from flexibility.py

    # region Manually Building A MoleculeSetup
    def add_atom(
        self,
        atom_index: int = None,
        overwrite: bool = False,
        pdbinfo: str or PDBAtomInfo = DEFAULT_PDBINFO,
        charge: float = DEFAULT_CHARGE,
        coord: np.ndarray = None,
        atomic_num: int = DEFAULT_ATOMIC_NUM,
        atom_type: str = DEFAULT_ATOM_TYPE,
        is_ignore: bool = DEFAULT_IS_IGNORE,
        graph: list[int] = None,
    ):
        """
        Adds an atom with all the specified attributes to the MoleculeSetup, either at the specified atom index, or by
        appending it to the internal list of atoms. Default values will be used for any attributes with unspecified
        values.

        Parameters
        ----------
        atom_index: int
            atom index in the MoleculeSetup
        overwrite: bool
            can we overwrite other atoms may be in the same atom index as this one
        pdbinfo: str
            pdb string for the atom
        coord: np.ndarray
            the atom's coordinates
        charge: float
            partial charge to be loaded for the atom
        atomic_num: int
            the atomic number of the atom
        atom_type: str
            TODO: needs info
        is_ignore: bool
            ignore flag for the atom
        graph: List[List[int]]

        Returns
        -------
        None

        Raises
        ------
        RuntimeException
            If the user tries to overwrite an existing atom without explicitly allowing overwrites.
        """
        # If atom index is specified and it would be trying to overwrite an existing atom in the atom list, raises a
        # Runtime Exception
        insert_disallowed = len(self.atoms) > atom_index and not overwrite
        if (
            atom_index is not None
            and insert_disallowed
            and not self.atoms[atom_index].is_dummy
        ):
            raise RuntimeError(
                "ADD_ATOM Error: the atom_index [%d] is already occupied (use 'overwrite' to force)"
            )

        # If atom index is not specified, appends the new atom to the end of the current atom list
        if atom_index is None:
            atom_index = len(self.atoms)

        # Inserts dummy atoms if a specified atom index is greater than the current length of the atom list
        while atom_index > len(self.atoms):
            self.atoms.append(Atom(len(self.atoms), is_dummy=True))

        # Creates and adds new atom to the atom list
        if coord is None:
            coord = deepcopy(DEFAULT_COORD)
        if graph is None:
            graph = deepcopy(DEFAULT_GRAPH)
        new_atom = Atom(
            atom_index,
            pdbinfo,
            charge,
            coord,
            atomic_num,
            atom_type,
            is_ignore,
            graph,
        )
        if atom_index < len(self.atoms):
            self.atoms[atom_index] = new_atom
            return
        self.atoms.append(new_atom)
        return

    def add_pseudoatom(
        self,
        pdbinfo: str or PDBAtomInfo = DEFAULT_PDBINFO,
        charge: float = DEFAULT_CHARGE,
        coord: np.ndarray = None,
        atom_type: str = DEFAULT_ATOM_TYPE,
        is_ignore: bool = DEFAULT_IS_IGNORE,
        anchor_list: list[int] = None,
        rotatable: bool = False,
        directional_vectors: list[int] = None,
    ):
        """
        Adds a pseudoatom with all the specified attributes to the MoleculeSetup. Default values will be used for any
        attributes with unspecified values. Multiple bonds can be specified to support the centroids of aromatic rings.
        If rotatable, makes the anchor atom rotatable to allow the pseudoatom movement.

        Parameters
        ----------
        pdbinfo: str
            PDB string for the pseudoatom.
        charge: float
            partial charge for the pseudoatom
        coord: np.ndarray
            the pseudoatom's coordinates
        atom_type: str
            TODO: needs info
        is_ignore: bool
            ignore flag for the pseudoatom
        anchor_list: list[int]
            a list of ints indicating the multiple bonds that can be specified as input
        rotatable: bool
            flag indicating if the anchor atom should be marked as rotatable to allow the pseudoatom movement.
        directional_vectors
            TODO: needs info

        Returns
        -------
        pseudoatom_index: int
            The atom_index of the added pseudoatom

        Raises
        ------
        RuntimeError:
            When the incorrect number of anchors of pseudoatoms are found in rigid_groups in the flexibility model

        """
        # Places the atom at the end of the atom list.
        pseudoatom_index = len(self.atoms)
        # Creates the atom and marks it as a pseudoatom
        if coord is None:
            coord = deepcopy(DEFAULT_COORD)
        new_pseudoatom = Atom(
            pseudoatom_index,
            pdbinfo=pdbinfo,
            charge=charge,
            coord=coord,
            atomic_num=self.PSEUDOATOM_ATOMIC_NUM,
            atom_type=atom_type,
            is_ignore=is_ignore,
            is_pseudo_atom=True,
        )
        self.atoms.append(new_pseudoatom)
        # Adds bonds for all of the provided anchor atoms
        if anchor_list is not None:
            for anchor in anchor_list:
                self.add_bond(pseudoatom_index, anchor, rotatable=rotatable)
        # Adds directional vectors [Check what this is used for/if this is used]
        if directional_vectors is not None:
            self._add_interaction_vectors(pseudoatom_index, directional_vectors)
        # If there are no specified anchor atoms,
        if not self.flexibility_model or not anchor_list:
            return pseudoatom_index
        # TODO: revise this logic
        # If there is a flexibility model in the MoleculeSetup, adds the psuedoatom to the flexibility model's rigid
        # group tracking
        rigid_groups_indices = []
        for anchor in anchor_list:
            for rigid_index, members in self.flexibility_model[
                "rigid_body_members"
            ].items():
                if anchor in members:
                    rigid_groups_indices.append(rigid_index)
        if len(rigid_groups_indices) != 1:
            raise RuntimeError(
                f"anchors of pseudo atom found in {len(rigid_groups_indices)} rigid_groups (must be 1)"
            )
        rigid_index = rigid_groups_indices[0]
        self.flexibility_model["rigid_body_members"][rigid_index].append(
            pseudoatom_index
        )
        # returns the pseudoatom index
        return pseudoatom_index

    def delete_atom(self, atom_index: int):
        """
        Clears the atom data at a specified atom index and replaces the atom with a dummy atom.

        Parameters
        ----------
        atom_index: int
            atom index to replace with a dummy atom

        Returns
        -------
        None
        """
        blank_atom = Atom(atom_index, is_dummy=True)
        self.atoms[atom_index] = blank_atom
        return

    def add_bond(
        self,
        atom_index_1: int,
        atom_index_2: int,
        rotatable: bool = DEFAULT_BOND_ROTATABLE,
    ) -> None:
        """
        Creates a bond and adds it to all the internal data structures where atom bonds are being tracked.

        Parameters
        ----------
        atom_index_1: int
            Atom index of one of the atoms in the bond
        atom_index_2: int
            Atom index of the other atom in the bond
        rotatable: bool
            Indicates whether the bond is rotatable

        Returns
        -------
        None

        Raises
        ------
        IndexError:
            When one or more the given bond atom indices do not exist in the MoleculeSetup
        """
        # Checks that both of the atom indices provided are valid indices, otherwise throws an error
        if len(self.atoms) <= atom_index_1 or len(self.atoms) <= atom_index_2:
            raise IndexError(
                "ADD_BOND: provided atom indices outside the range of atoms currently in MoleculeSetup"
            )
        # Adds each atom to the other's bond graph
        if atom_index_2 not in self.atoms[atom_index_1].graph:
            self.atoms[atom_index_1].graph.append(atom_index_2)
        if atom_index_1 not in self.atoms[atom_index_2].graph:
            self.atoms[atom_index_2].graph.append(atom_index_1)
        # Creates new bond object and uses its internal canonical bond id to add it to MoleculeSetup bond tracking.
        new_bond = Bond(atom_index_1, atom_index_2, rotatable)
        self.bond_info[new_bond.canon_id] = new_bond
        return

    def delete_bond(self, atom_index_1: int, atom_index_2: int):
        """
        Deletes a bond from the molecule setup.

        Parameters
        ----------
        atom_index_1: int
            The atom index of one of the atoms in the bond to delete
        atom_index_2: int
            The atom index of the other atom in the bond to delete

        Returns
        -------
        None
        """
        # Gets canon bond id for the bond to delete
        canon_bond_id = Bond.get_bond_id(atom_index_1, atom_index_2)
        # Deletes the bond from the internal bond table
        del self.bond_info[canon_bond_id]
        # Removes the bond from each atom's graph
        self.atoms[atom_index_1].graph.remove(atom_index_2)
        self.atoms[atom_index_2].graph.remove(atom_index_1)
        return

    def add_rotamers(
        self, index_list: list[(int, int, int, int)], angle_list: np.ndarray
    ):
        """
        Adds rotamers to the internal record of rotamers.

        Parameters
        ----------
        index_list: list[(int, int, int, int)]
        angle_list: np.ndarray

        Returns
        -------
        None
        """
        # It's unclear how this will work without the coordinates in the moleculesetup. food for thought.
        # TODO: address issues with the lack of coords and add detail in function comment
        rotamers = {}
        for (idx1, idx2, idx3, idx4), angle in zip(index_list, angle_list):
            bond_id = Bond.get_bond_id(idx2, idx3)
            if bond_id in rotamers:
                raise RuntimeError("repeated bond %d-$d" % bond_id)
            if not self.bond_info[bond_id].rotatable:
                raise RuntimeError(
                    "trying to add rotamer for non rotatable bond %d-%d" % bond_id
                )
            # d0 = calcDihedral(xyz[i1], xyz[i2], xyz[i3], xyz[i4])
            dihedral = 0  # TODO: fix this
            rotamers[bond_id] = angle - dihedral
        self.rotamers.append(rotamers)
        return

    def delete_rotamers(
        self,
        bond_id_list: list[tuple] = None,
        index_list: list[(int, int, int, int)] = None,
    ):
        """
        Deletes rotamers from the internal list of rotamers, either by using bond ids or by generating bond ids from a
        list of input indices.

        Parameters
        ----------
        bond_id_list: list[tuple]
        index_list: list[(int, int, int, int)]

        Returns
        -------
        None
        """
        # loops through the index list, generates bond ids from the provided indices and adds them to the bond_id_list
        if index_list is not None:
            for idx1, idx2, idx3, idx4 in index_list:
                bond_id = Bond.get_bond_id(idx2, idx3)
                bond_id_list.append(bond_id)
        # deletes all bond_ids in bond_id_list from self.rotamers
        if bond_id_list is not None:
            for bond_id in bond_id_list:
                if bond_id in self.rotamers:
                    del self.rotamers[bond_id]
        return

    def _add_interaction_vectors(self, atom_index: int, vector_list: list[np.array]):
        """
        Adds input vector list to the list of directional interaction vectors for the specified atom.

        Parameters
        ----------
        atom_index: int
            index of the atom to add the vectors to
        vector_list: list[np.array]
            a list of directional interaction vectors

        Returns
        -------
        None

        Raises
        ------
        IndexError
            if the specified atom index does not exist or is a dummy atom.
        """
        if atom_index > len(self.atoms) or self.atoms[atom_index].is_dummy:
            raise IndexError(
                "INTERACTION_VECTORS: provided atom index is out of range or is a dummy atom."
            )
        for vector in vector_list:
            self.atoms[atom_index].interaction_vectors.append(vector)
        return

    @property
    def true_atom_count(self):
        """
        Counts the number of atoms in the MoleculeSetup that are not pseudo_atoms or marked as dummy atoms

        Returns
        -------
        count: int
            The number of atoms currently in the MoleculeSetup that are not dummy atoms or pseudo_atoms.
        """
        count = 0
        for atom in self.atoms:
            if not atom.is_pseudo_atom and not atom.is_dummy:
                count += 1
        return count

    # this might void the graph connections and everything in bonds, might need to add a dict of the changes we're
    # making and then use that save this for a future push.
    def clean_atoms(self, remove_pseudoatoms: bool = False):
        """
        Cleans dummy and potentially also pseudoatoms from the MoleculeSetup so only true atoms remain. Note that this
        is pretty slow and should not be done often.

        Parameters
        ----------
        remove_pseudoatoms: bool
            Indicates if we want to remove all the pseudoatoms from the MoleculeSetup.

        Returns
        -------
        The number of atoms removed from the MoleculeSetup.
        """
        new_atoms = []
        removed_atom_count = 0
        atom_index_mapping = {}
        for atom in self.atoms:
            if remove_pseudoatoms and atom.is_pseudo_atom:
                removed_atom_count += 1
                continue
            if atom.is_dummy:
                removed_atom_count += 1
                continue
            atom.index = atom.index - removed_atom_count
            new_atoms.append(atom)
        self.atoms = new_atoms
        if remove_pseudoatoms:
            self.pseudoatom_count = 0
        return removed_atom_count

    # endregion

    # region Getters and Setters

    def get_pdbinfo(self, atom_index: int):
        """
        Retrieves the PDB Info string for the atom with the specified atom index.

        Parameters
        ----------
        atom_index: int
            Atom index to retrieve data for.

        Returns
        -------
        pdbinfo: str
            A string containing the pdb information for the atom

        Raises
        ------
        IndexError:
            When the provided atom index does not exist in the MoleculeSetup or the atom index does not contain
            data.
        """
        if atom_index > len(self.atoms) or self.atoms[atom_index].is_dummy:
            raise IndexError(
                "GET_PDBINFO: provided atom index is out of range or is a dummy atom"
            )
        return self.atoms[atom_index].pdbinfo

    def get_charge(self, atom_index: int):
        """
        Retrieves the partial charge for the atom with the specified atom index.

        Parameters
        ----------
        atom_index: int
            Atom index to retrieve data for.

        Returns
        -------
        charge: float
            The charge associated with the atom

        Raises
        ------
        IndexError:
            When the provided atom index does not exist in the MoleculeSetup or the atom index does not contain
            data.
        """
        if atom_index > len(self.atoms) or self.atoms[atom_index].is_dummy:
            raise IndexError(
                "GET_CHARGE: provided atom index is out of range or is a dummy atom"
            )
        return self.atoms[atom_index].charge

    def get_coord(self, atom_index: int):
        """
        Retrieves the coordinates for the atom with the specified atom index.

        Parameters
        ----------
        atom_index: int
            Atom index to retrieve data for.

        Returns
        -------
        coord: np.ndarray
            The coordinates associated with the atom.

        Raises
        ------
        IndexError:
            When the provided atom index does not exist in the MoleculeSetup or the atom index does not contain
            data.
        """
        if atom_index > len(self.atoms) or self.atoms[atom_index].is_dummy:
            raise IndexError(
                "GET_CHARGE: provided atom index is out of range or is a dummy atom"
            )
        return self.atoms[atom_index].coord

    def get_atomic_num(self, atom_index: int):
        """
        Retrieves the atomic number for the atom with the specified atom index.

        Parameters
        ----------
        atom_index: int
            Atom index to retrieve data for.

        Returns
        -------
        atomic_num: int
            The atomic number associated with an atom.

        Raises
        ------
        IndexError:
            When the provided atom index does not exist in the MoleculeSetup or the atom index does not contain
            data.
        """
        if atom_index > len(self.atoms) or self.atoms[atom_index].is_dummy:
            raise IndexError(
                "GET_ATOMIC_NUM: provided atom index is out of range or is a dummy atom"
            )
        return self.atoms[atom_index].atomic_num

    def get_atom_type(self, atom_index: int):
        """
        Retrieves the atom type for the atom with the specified atom index.

        Parameters
        ----------
        atom_index: int
            Atom index to retrieve data for.

        Returns
        -------
        charge: str
            The atom index associated with the atom

        Raises
        ------
        IndexError:
            When the provided atom index does not exist in the MoleculeSetup or the atom index does not contain data.
        """
        if atom_index > len(self.atoms) or self.atoms[atom_index].is_dummy:
            raise IndexError(
                "GET_ATOM_TYPE: provided atom index is out of range or is a dummy atom"
            )
        return self.atoms[atom_index].atom_type

    def set_atom_type(self, atom_index: int, atom_type: str) -> None:
        """
        Sets the atom type for the atom with the specified atom index.

        Parameters
        ----------
        atom_index: int
            Atom index to set atom_type for.
        atom_type
            Atom type string to set.

        Returns
        -------
        None

        Raises
        ------
        IndexError:
            When the provided atom index does not exist in the MoleculeSetup or the atom index does not contain data.
        """
        if atom_index > len(self.atoms) or self.atoms[atom_index].is_dummy:
            raise IndexError(
                "SET_ATOM_TYPE: provided atom index is out of range or is a dummy atom"
            )
        self.atoms[atom_index].atom_type = atom_type
        return

    def set_atom_type_from_uniq_atom_params(
        self, uniq_atom_params: UniqAtomParams, prefix: str
    ):
        """
        Uses a UniqAtomParams object to set the atom_type attribute for atoms in the Molecule Setup object. Adds the specified prefix
        to each of the atom_type attributes pulled from UniqAtomParams.

        Parameters
        ----------
        uniq_atom_params: UniqAtomParams
            A uniq atom params object to extract atom_type from
        prefix: string
            A prefix to be appended to all the atom_type attributes

        Returns
        -------
        None
        """
        # Gets a mapping from parameter indices in atom_params to those in uniq_atom_params
        parameter_indices = uniq_atom_params.get_indices_from_atom_params(
            self.atom_params
        )
        # Checks that we have the correct number of retrieved indices.
        if len(parameter_indices) != len(self.atoms):
            raise RuntimeError(
                "Number of parameters ({len(parameter_indices)}) not equal to number of atoms in Molecule Setup ({len(self.atom_type)})"
            )
        # Loops through the indices in parameter indices and sets atom types with the input prefix
        for i, j in enumerate(parameter_indices):
            self.atom_type[i] = f"{prefix}{j}"
        return None

    def get_is_ignore(self, atom_index: int):
        """
        Retrieves the is_ignore boolean for the atom with the specified atom index.

        Parameters
        ----------
        atom_index: int
            Atom index to retrieve data for.

        Returns
        -------
        is_ignore: bool
            Indicates whether a particular atom should be ignored

        Raises
        ------
        IndexError:
            When the provided atom index does not exist in the MoleculeSetup or the atom index does not contain
            data.
        """
        if atom_index > len(self.atoms) or self.atoms[atom_index].is_dummy:
            raise IndexError(
                "GET_IS_IGNORE: provided atom index is out of range or is a dummy atom"
            )
        return self.atoms[atom_index].is_ignore

    def get_neighbors(self, atom_index: int):
        """
        Retrieves the partial charge for the atom with the specified atom index.

        Parameters
        ----------
        atom_index: int
            Atom index to retrieve data for.

        Returns
        -------
        graph: list[int]
            The graph of the atoms connections to other atoms.

        Raises
        ------
        IndexError:
            When the provided atom index does not exist in the MoleculeSetup or the atom index does not contain
            data.
        """
        if atom_index > len(self.atoms) or self.atoms[atom_index].is_dummy:
            raise IndexError(
                "GET_GRAPH: provided atom index is out of range or is a dummy atom"
            )
        return self.atoms[atom_index].graph

    def get_interaction_vectors(self, atom_index: int):
        if atom_index > len(self.atoms) or self.atoms[atom_index].is_dummy:
            raise IndexError(
                "GET_INTERACTION_VECTORS: provided atom index is out of range or is a dummy atom"
            )
        return self.atoms[atom_index].interaction_vectors

    # endregion

    def merge_terminal_atoms(self, indices) -> None:
        """
        Primarily for merging hydrogens, but will merge the data for any atom or pseudoatom that is bonded to only one
        other atom.

        Parameters
        ----------
        indices: list
            A list of indices to merge

        Returns
        -------
        None
        """
        for index in indices:
            if len(self.get_neighbors(index)) != 1:
                msg = "Atempted to merge atom %d with %d neighbors. "
                msg += "Only atoms with one neighbor can be merged."
                msg = msg % (index + 1, self.get_neighbors(index))
                raise RuntimeError(msg)
            neighbor_index = self.get_neighbors(index)[0]
            self.atoms[neighbor_index].charge += self.get_charge(index)
            self.atoms[index].charge = 0.0
            self.atoms[index].is_ignore = True
        return

    # NOTE: This is a candidate for moving to utils
    @staticmethod
    def get_bonds_in_ring(ring: tuple) -> list[tuple]:
        """
        Takes as input a tuple of atom indices corresponding to atoms in a ring and returns a list of all the bonds ids
        in the ring.

        Parameters
        ----------
        ring: tuple
            A list of atom indices of the atoms in a ring.

        Returns
        -------
        A list of canonical bond id tuples for the bonds in the ring.
        """
        bonds = []
        num_indices = len(ring)
        for i in range(num_indices):
            bond = (ring[i], ring[(i + 1) % num_indices])
            bond = Bond.get_bond_id(bond[0], bond[1])
            bonds.append(bond)
        return bonds

    def _recursive_graph_walk(
        self, idx: int, collected: list[int] = None, exclude: list[int] = None
    ):
        """
        Recursively walks through a molecular graph and returns bond-connected subgroups.

        Parameters
        ----------
        idx: int
            atom index to start the recursive walk from
        collected: list[int]
            a list of connected subgroups
        exclude: list[int]
            a list of atom indices to exclude from the final walk.

        Returns
        -------
        A list of ints indicating the subgroups that are bond-connected.
        """
        if collected is None:
            collected = []
        if exclude is None:
            exclude = []
        for neighbor in self.get_neighbors(idx):
            if neighbor in collected or neighbor in exclude:
                continue
            collected.append(neighbor)
            self._recursive_graph_walk(neighbor, collected, exclude)
        return collected

    def write_coord_string(self) -> str:
        """
        Constructs and returns a string of all atom and pseudoatom elements and coordinates.

        Returns
        -------
        A string of all atom and pseudoatom elements and coordinates.
        """
        n = len(self.atoms)
        output_string = "%d\n\n" % n
        for index in range(n):
            element = "Ne"
            if self.atoms[index].is_dummy:
                continue
            if not self.atoms[index].is_pseudo_atom:
                element = utils.mini_periodic_table[self.atoms[index].atomic_num]
            x, y, z = self.atoms[index].coord
            output_string += "%3s %12.6f %12.6f %12.6f\n" % (element, x, y, z)
        return output_string

    def show(self) -> None:
        """
        Legacy function to print the contents of a MoleculeSetup in a human-readable format.

        Returns
        -------
        None
        """
        total_charge = 0

        print("Molecule Setup\n")
        print(
            "==============[ ATOMS ]==================================================="
        )
        print("idx  |          coords            | charge |ign| atype    | connections")
        print(
            "-----+----------------------------+--------+---+----------+--------------- . . . "
        )
        for atom in self.atoms:
            print(
                "% 4d | % 8.3f % 8.3f % 8.3f | % 1.3f | %d"
                % (
                    atom.index,
                    atom.coord[0],
                    atom.coord[1],
                    atom.coord[2],
                    atom.charge,
                    atom.is_ignore,
                ),
                "| % -8s |" % atom.atom_type,
                atom.graph,
            )
            total_charge += atom.charge
        print(
            "-----+----------------------------+--------+---+----------+--------------- . . . "
        )
        print("  TOT CHARGE: %3.3f" % total_charge)

        print("\n==============[ BONDS ]================")
        # For sanity users, we won't show those keys for now
        keys_to_not_show = ["type"]
        for bond_id, bond in list(self.bond_info.items()):
            t = ", ".join(
                "%s: %s" % (i, j)
                for i, j in bond.__dict__.items()
                if i not in keys_to_not_show
            )
            print("% 8s - " % str(bond_id), t)
        print("")
        return

    def to_json(self):
        """
        Converts MoleculeSetup object to a JSON string.

        Returns
        -------
        A JSON string representation of the MoleculeSetup object.
        """
        return json.dumps(self, cls=MoleculeSetupEncoder)

    @staticmethod
    def from_json(obj):
        """
        Takes an object and attempts to decode it into a MoleculeSetup object.

        Parameters
        ----------
        obj: Object
            This can be any object, but it should be a dictionary generated by deserializing a JSON of a MoleculeSetup
            object.

        Returns
        -------
        If the input is a dictionary corresponding to a MoleculeSetup, will return a MoleculeSetup with data
        populated from the dictionary. Otherwise, returns the input object.
        """
        # if the input object is not a dict, we know that it will not be parsable and is unlikely to be usable or
        # safe data, so we should ignore it.
        if type(obj) is not dict:
            return obj

        # checks that all the keys we expect are in the object dictionary as a safety measure. Allows other keys
        # to exist in case this is being called by a subclass of MoleculeSetup that may have more keys in its dict.
        expected_molsetup_keys = {
            "name",
            "is_sidechain",
            "pseudoatom_count",
            "atoms",
            "bond_info",
            "rings",
            "ring_closure_info",
            "rotamers",
            "atom_params",
            "restraints",
            "flexibility_model",
        }
        for key in expected_molsetup_keys:
            if key not in obj.keys():
                return obj

        # Constructs a MoleculeSetup object and restores the expected attributes
        name = obj["name"]
        is_sidechain = obj["is_sidechain"]
        molsetup = MoleculeSetup(name, is_sidechain)
        molsetup.pseudoatom_count = obj["pseudoatom_count"]
        molsetup.atoms = [Atom.from_json(x) for x in obj["atoms"]]
        molsetup.bond_info = {
            string_to_tuple(k, int): Bond.from_json(v)
            for k, v in obj["bond_info"].items()
        }
        molsetup.rings = {
            string_to_tuple(k, int): Ring.from_json(v) for k, v in obj["rings"].items()
        }
        molsetup.ring_closure_info = RingClosureInfo(
            obj["ring_closure_info"]["bonds_removed"],
            obj["ring_closure_info"]["pseudos_by_atom"],
        )
        molsetup.rotamers = obj["rotamers"]
        molsetup.atom_params = obj["atom_params"]
        molsetup.restraints = [Restraint.from_json(x) for x in obj["restraints"]]
        molsetup.flexibility_model = obj["flexibility_model"]
        if "rigid_body_connectivity" in molsetup.flexibility_model:
            tuples_rigid_body_connectivity = {
                string_to_tuple(k, int): string_to_tuple(v)
                for k, v in molsetup.flexibility_model[
                    "rigid_body_connectivity"
                ].items()
            }
            molsetup.flexibility_model["rigid_body_connectivity"] = (
                tuples_rigid_body_connectivity
            )
        if "rigid_body_graph" in molsetup.flexibility_model:
            molsetup.flexibility_model["rigid_body_graph"] = {
                int(k): v
                for k, v in molsetup.flexibility_model["rigid_body_graph"].items()
            }
        if "rigid_body_members" in molsetup.flexibility_model:
            molsetup.flexibility_model["rigid_body_members"] = {
                int(k): v
                for k, v in molsetup.flexibility_model["rigid_body_members"].items()
            }
        if "rigid_index_by_atom" in molsetup.flexibility_model:
            molsetup.flexibility_model["rigid_index_by_atom"] = {
                int(k): v
                for k, v in molsetup.flexibility_model["rigid_index_by_atom"].items()
            }
        return molsetup


# region External Toolkit Support
class MoleculeSetupExternalToolkit(ABC):
    """
    Additional functions and requirements to extend the MoleculeSetup class in order to use it with  external toolkits
    such as RDKit and OpenBabel.

    Required Attributes
    -------------------
    dihedral_interactions: list
        A list of fourier series [add detail]
    """

    @staticmethod
    def are_fourier_series_identical(series1: list, series2: list) -> bool:
        """
        Compares two fourier series represented as lists of dictionaries.

        Parameters
        ----------
        series1: list[dict]
            The first fourier series to compare.
        series2: list[dict]
            The second fourier series to compare.

        Returns
        -------
        A bool indicicating whether the fourier series are equal.
        """
        # Gets the indices of both series by periodicity and checks for equality
        index_by_periodicity1 = {
            series1[index]["periodicity"]: index for index in range(len(series1))
        }
        index_by_periodicity2 = {
            series2[index]["periodicity"]: index for index in range(len(series2))
        }
        if index_by_periodicity1 != index_by_periodicity2:
            return False
        # After establishing equality of the indices, loops through periodicity abd checks that the values stored in
        # each fourier series dictionary are equal.
        for periodicity in index_by_periodicity1:
            index1 = index_by_periodicity1[periodicity]
            index2 = index_by_periodicity2[periodicity]
            for key in ["k", "phase", "periodicity"]:
                if series1[index1][key] != series2[index2][key]:
                    return False
        return True

    def add_dihedral_interaction(self, fourier_series):
        """
        Adds a safe copy of the input fourier series to the dihedral_interactions list if the fourier series is not
        already in the list.

        Parameters
        ----------
        fourier_series: list[dict]

        Returns
        -------
        index: int
            The index of the input fourier series in the dihedral interactions list.
        """
        index = 0
        for existing_fs in self.dihedral_interactions:
            if self.are_fourier_series_identical(existing_fs, fourier_series):
                return index
            index += 1
        safe_copy = json.loads(json.dumps(fourier_series))
        self.dihedral_interactions.append(safe_copy)
        return index

    @abstractmethod
    def init_atom(self, assign_charges, coords):
        pass

    @abstractmethod
    def init_bond(self):
        pass

    @abstractmethod
    def get_mol_name(self):
        pass

    @abstractmethod
    def find_pattern(self, smarts: str):
        pass

    @abstractmethod
    def get_smiles_and_order(self):
        pass

    pass


class RDKitMoleculeSetup(MoleculeSetup, MoleculeSetupExternalToolkit):
    """
    Subclass of MoleculeSetup, used to represent MoleculeSetup objects working with RDKit objects

    Attributes
    ----------
    mol : rdkit.Chem.rdchem.Mol
        An RDKit Mol object to base the Molecule Setup on.
    modified_atom_positions: list
        List of dictionaries where keys are atom indices, Used to store sets of coordinates, e.g. docked poses, as
        dictionaries indexed by the atom index, because not all atoms need to have new coordinates specified.
        Unspecified hydrogen positions bonded to modified heavy atom positions are to be calculated "on-the-fly".
    dihedral_interactions: list[]
        A list of unique fourier_series, each of which are represented as a list of dictionaries.
    dihedral_partaking_atoms: dict()
        a mapping from tuples of atom indices to the indices in dihedral_interactions
    dihedral_labels: dict()
        a mapping from tuples of atom indices to dihedral labels
    atom_to_ring_id: dict()
        mapping of atom index to ring id of each atom belonging to the ring
    ring_corners: dict()
        unclear what this is a mapping of, but is used to store corner flexibility for the rings
    rmsd_symmetry_indices: tuple
        Tuples of the indices of the molecule's atoms that match a substructure query. needs info.

    Methods
    -------
    from_mol()
        constructor for the RDKitMoleculeSetup object (consider adapting to init?)
    """

    def __init__(self, name: str = None, is_sidechain: bool = False):
        super().__init__(name, is_sidechain)
        self.mol = None
        self.modified_atom_positions = []
        self.dihedral_interactions: list[dict] = []
        self.dihedral_partaking_atoms: dict = {}
        self.dihedral_labels: dict = {}
        self.atom_to_ring_id = {}
        self.ring_corners = {}
        self.rmsd_symmetry_indices = ()

    def copy(self):
        """
        Returns a copy of the current RDKitMoleculeSetup.
        """
        newsetup = RDKitMoleculeSetup()
        for key, value in self.__dict__.items():
            if key != "mol":
                newsetup.__dict__[key] = deepcopy(value)
        newsetup.mol = Chem.Mol(self.mol)  # not sure how deep of a copy this is
        return newsetup

    @classmethod
    def from_mol(
        cls,
        mol: Chem.Mol,
        keep_chorded_rings: bool = False,
        keep_equivalent_rings: bool = False,
        assign_charges: bool = True,
        conformer_id: int = -1,
    ):
        """

        Parameters
        ----------
        mol: rdkit.Chem.rdchem.Mol
            RDKit Mol object to build the RDKitMoleculeSetup from.
        keep_chorded_rings: bool
        keep_equivalent_rings: bool
        assign_charges: bool
        conformer_id: int

        Returns
        -------
        molsetup: RDKitMoleculeSetup
            A populated RDKitMoleculeSetup object

        Raises
        ------
        ValueError:
            If the RDKit Mol has implicit Hydrogens or if there are no conformers for the given RDKit Mol
        """
        # Checks if the input molecule is valid
        if cls.has_implicit_hydrogens(mol):
            raise ValueError("RDKit molecule has implicit Hs. Need explicit Hs.")
        if mol.GetNumConformers() == 0:
            raise ValueError(
                "RDKit molecule does not have a conformer. Need 3D coordinates."
            )

        # Gets the RDKit Conformer that we are going to load into the molecule setup
        rdkit_conformer = mol.GetConformer(conformer_id)
        if not rdkit_conformer.Is3D():
            warnings.warn(
                "RDKit molecule not labeled as 3D. This warning won't show again."
            )
            RDKitMoleculeSetup.warned_not3D = True
        if mol.GetNumConformers() > 1 and conformer_id == -1:
            msg = "RDKit molecule has multiple conformers. Considering only the first one."
            print(msg, file=sys.stderr)
        if len(Chem.GetMolFrags(mol)) != 1:
            raise ValueError(f"RDKit molecule has {len(Chem.GetMolFrags(mol))} fragments. Must have 1.")
        if mol.HasQuery():
            raise ValueError(f"RDKit molecule has query. Check exotic fields (atom or bond) in SDF.")

        # Creating and populating the molecule setup with properties from RDKit as well as calculated values from our
        # functions
        molsetup = cls()
        molsetup.mol = mol
        molsetup.atom_true_count = molsetup.get_num_mol_atoms()
        molsetup.name = molsetup.get_mol_name()
        coords = rdkit_conformer.GetPositions()
        molsetup.init_atom(assign_charges, coords)
        molsetup.init_bond()
        molsetup.perceive_rings(keep_chorded_rings, keep_equivalent_rings)
        molsetup.rmsd_symmetry_indices = cls.get_symmetries_for_rmsd(mol)

        # to store sets of coordinates, e.g. docked poses, as dictionaries indexed by
        # the atom index, because not all atoms need to have new coordinates specified
        # Unspecified hydrogen positions bonded to modified heavy atom positions
        # are to be calculated "on-the-fly".
        molsetup.modified_atom_positions = (
            []
        )  # list of dictionaries where keys are atom indices

        return molsetup

    @staticmethod
    def remove_elements(mol, to_rm=(12, 20, 25, 26, 30)):
        idx_to_rm = {}
        neigh_idx_to_nr_h = {}
        rm_to_neigh = {}
        for atom in mol.GetAtoms():
            if atom.GetAtomicNum() in to_rm:
                idx_to_rm[atom.GetIdx()] = atom.GetFormalCharge()
                rm_to_neigh[atom.GetIdx()] = set()
                for neigh in atom.GetNeighbors():
                    n = neigh.GetNumExplicitHs()
                    neigh_idx_to_nr_h[neigh.GetIdx()] = n
                    rm_to_neigh[atom.GetIdx()].add(neigh.GetIdx())
        if not idx_to_rm:
            return Chem.Mol(mol), idx_to_rm, rm_to_neigh
        rwmol = Chem.EditableMol(mol)
        for idx in sorted(idx_to_rm, reverse=True):
            rwmol.RemoveAtom(idx)
        mol = rwmol.GetMol()
        for idx in neigh_idx_to_nr_h:
            n = neigh_idx_to_nr_h[idx]
            newidx = idx - sum([i < idx for i in idx_to_rm]) 
            mol.GetAtomWithIdx(newidx).SetNumExplicitHs(n + 1)
        mol.UpdatePropertyCache()
        Chem.SanitizeMol(mol)
        mol = Chem.AddHs(mol)
        return mol, idx_to_rm, rm_to_neigh
         

    def init_atom(self, assign_charges: bool, coords: list[np.ndarray]):
        """
        Generates information about the atoms in an RDKit Mol and adds them to an RDKitMoleculeSetup.

        Parameters
        ----------
        assign_charges: bool
            Indicates whether we should extract/generate charges.
        coords: list[np.ndarray]
            Atom coordinates for the RDKit Mol.

        Returns
        -------
        None
        """
        # extract/generate charges
        if assign_charges:
            things = self.remove_elements(self.mol)
            copy_mol, idx_rm_to_formal_charge, rm_to_neigh = things
            for atom in copy_mol.GetAtoms():
                if atom.GetAtomicNum() == 34:
                    atom.SetAtomicNum(16)
            rdPartialCharges.ComputeGasteigerCharges(copy_mol)
            charges = [a.GetDoubleProp("_GasteigerCharge") for a in copy_mol.GetAtoms()]
            if idx_rm_to_formal_charge:
                ok_charges = charges.copy()
                for i in sorted(idx_rm_to_formal_charge, reverse=True):
                    ok_charges.insert(i, 0.0)
                nr_rm = len(idx_rm_to_formal_charge)
                nr_added_h = copy_mol.GetNumAtoms() - self.mol.GetNumAtoms() + nr_rm
                ok_charges = ok_charges[:len(ok_charges)-nr_added_h]
                # print(f"{nr_added_h=}")
                # print(f"{nr_rm=}")
                # print(f"{idx_rm_to_formal_charge=}")
                # print(f"{len(charges)=}")
                # print(f"{len(ok_charges)=}")
                # print(f"{copy_mol.GetNumAtoms()=}")
                # print(f"{self.mol.GetNumAtoms()=}")
                chrg_by_heavy_atom = {}
                for i in range(nr_added_h):
                    added_H_idx = self.mol.GetNumAtoms() + i - nr_rm
                    # print(f"{added_H_idx=}")
                    neighs = copy_mol.GetAtomWithIdx(added_H_idx).GetNeighbors()
                    if len(neighs) != 1:
                        raise RuntimeError("H should have 1 neighbor")
                    if neighs[0].GetIdx() in chrg_by_heavy_atom:
                        raise RuntimeError("expected only 1 added H per heavy atom, maybe deleted element had double bond to this heavy atom")
                    chrg_by_heavy_atom[neighs[0].GetIdx()] = charges[added_H_idx]
                # print(f"{chrg_by_heavy_atom=}")
                for i, neighs in rm_to_neigh.items():
                    # print(f"{i=}, {neighs=}")
                    ok_charges[i] += idx_rm_to_formal_charge[i]
                    for idx in neighs:
                        newidx = idx - sum([i <= idx for i in idx_rm_to_formal_charge]) 
                        # print(f"{idx=} {newidx=}")
                        ok_charges[i] += chrg_by_heavy_atom[newidx]
                charges = ok_charges
        else:
            charges = [0.0] * self.mol.GetNumAtoms()
        # register atom
        for a in self.mol.GetAtoms():
            idx = a.GetIdx()
            self.add_atom(
                atom_index=idx,
                pdbinfo=rdkitutils.getPdbInfoNoNull(a),
                charge=charges[idx],
                coord=coords[idx],
                atomic_num=a.GetAtomicNum(),
                is_ignore=False,
            )

    def init_bond(self):
        """
        Uses the RDKit mol to initialize bond info for the RDKitMoleculeSetup

        Returns
        -------
        None
        """
        for b in self.mol.GetBonds():
            idx1 = b.GetBeginAtomIdx()
            idx2 = b.GetEndAtomIdx()
            rotatable = int(b.GetBondType()) == 1
            self.add_bond(idx1, idx2, rotatable=rotatable)

    def find_pattern(self, smarts: str):
        """
        Given a SMARTS pattern, finds substruct matches in the molecule.

        Parameters
        ----------
        smarts:
            A SMARTS string to find in the RDKit Mol object

        Returns
        -------
        The substruct matches in the RDKit Mol for the given SMARTS.
        """
        p = Chem.MolFromSmarts(smarts)
        return self.mol.GetSubstructMatches(p)

    def get_mol_name(self):
        """
        Gets the RDKit Mol's name from self.mol.

        Returns
        -------
        If the mol has a name, returns the name property.
        """
        if self.mol.HasProp("_Name"):
            return self.mol.GetProp("_Name")
        else:
            return None

    # TODO: Add more inline comments and clean up this function
    def get_smiles_and_order(self):
        """
        Returns the SMILES string and the mapping between atom indices in the SMILES and self.molof an atom after
        running RDKit's RemoveHs function.

        Returns
        -------
        smiles:
        order:
        """
        mol_no_ignore = self.mol

        # 3D SDF files written by other toolkits (OEChem, ChemAxon)
        # seem to not include the chiral flag in the bonds block, only in
        # the atoms block. RDKit ignores the atoms chiral flag as per the
        # spec. When reading SDF (e.g. from PubChem/PDB),
        # we may need to have RDKit assign stereo from coordinates, see:
        # https://sourceforge.net/p/rdkit/mailman/message/34399371/
        ps = Chem.RemoveHsParameters()
        # a user reported PDBbind Mol Blocks to have hcount=1 for Hs,
        # which adds a query to the RDKit H atom and then Chem.RemoveHs
        # does not remove Hs with queries by default
        # https://github.com/forlilab/Meeko/issues/62
        # https://github.com/rdkit/rdkit/issues/6615
        ps.removeWithQuery = True
        mol_noH = Chem.RemoveHs(mol_no_ignore, ps)  # imines (=NH) may become chiral
        # stereo imines [H]/N=C keep [H] after RemoveHs()
        # H isotopes also kept after RemoveHs()
        atomic_num_mol_noH = [atom.GetAtomicNum() for atom in mol_noH.GetAtoms()]
        noH_to_H = []
        parents_of_hs = {}
        for index, atom in enumerate(mol_no_ignore.GetAtoms()):
            if atom.GetAtomicNum() == 1:
                continue
            for i in range(len(noH_to_H), len(atomic_num_mol_noH)):
                if atomic_num_mol_noH[i] > 1:
                    break
                h_atom = mol_noH.GetAtomWithIdx(len(noH_to_H))
                assert h_atom.GetAtomicNum() == 1
                neighbors = h_atom.GetNeighbors()
                assert len(neighbors) == 1
                parents_of_hs[len(noH_to_H)] = neighbors[0].GetIdx()
                noH_to_H.append("H")
            noH_to_H.append(index)
        extra_hydrogens = len(atomic_num_mol_noH) - len(noH_to_H)
        if extra_hydrogens > 0:
            assert set(atomic_num_mol_noH[len(noH_to_H) :]) == {1}
        for i in range(extra_hydrogens):
            h_atom = mol_noH.GetAtomWithIdx(len(noH_to_H))
            assert h_atom.GetAtomicNum() == 1
            neighbors = h_atom.GetNeighbors()
            assert len(neighbors) == 1
            parents_of_hs[len(noH_to_H)] = neighbors[0].GetIdx()
            noH_to_H.append("H")

        # noH_to_H has the same length as the number of atoms in mol_noH
        # and each value is:
        #    - the index of the corresponding atom in mol, if value is integer
        #    - an hydrogen, if value is "H"
        # now, we need to replace those "H" with integers
        # "H" occur with stereo imine (e.g. [H]/N=C) and heavy Hs (e.g. [2H])
        hs_by_parent = {}
        for hidx, pidx in parents_of_hs.items():
            hs_by_parent.setdefault(pidx, [])
            hs_by_parent[pidx].append(hidx)
        for pidx, hidxs in hs_by_parent.items():
            siblings_of_h = [
                atom
                for atom in mol_no_ignore.GetAtomWithIdx(noH_to_H[pidx]).GetNeighbors()
                if atom.GetAtomicNum() == 1
            ]
            sortidx = [
                i
                for i, j in sorted(
                    list(enumerate(siblings_of_h)), key=lambda x: x[1].GetIdx()
                )
            ]
            if len(hidxs) == len(siblings_of_h):
                # This is the easy case, just map H to each other in the order they appear
                for i, hidx in enumerate(hidxs):
                    noH_to_H[hidx] = siblings_of_h[sortidx[i]].GetIdx()
            elif len(hidxs) < len(siblings_of_h):
                # check hydrogen isotopes
                sibling_isotopes = [
                    siblings_of_h[sortidx[i]].GetIsotope()
                    for i in range(len(siblings_of_h))
                ]
                molnoH_isotopes = [mol_noH.GetAtomWithIdx(hidx) for hidx in hidxs]
                matches = []
                for i, sibling_isotope in enumerate(sibling_isotopes):
                    for hidx in hidxs[len(matches) :]:
                        if mol_noH.GetAtomWithIdx(hidx).GetIsotope() == sibling_isotope:
                            matches.append(i)
                            break
                if len(matches) != len(hidxs):
                    raise RuntimeError(
                        "Number of matched isotopes %d differs from query Hs: %d"
                        % (len(matches), len(hidxs))
                    )
                for hidx, i in zip(hidxs, matches):
                    noH_to_H[hidx] = siblings_of_h[sortidx[i]].GetIdx()
            else:
                raise RuntimeError(
                    "nr of Hs in mol_noH bonded to an atom exceeds nr of Hs in mol_no_ignore"
                )

        smiles = Chem.MolToSmiles(mol_noH)
        order_string = mol_noH.GetProp("_smilesAtomOutputOrder")
        order_string = order_string.replace(",]", "]")  # remove trailing comma
        order = json.loads(order_string)  # mol_noH to smiles
        order = list(np.argsort(order))
        order = {noH_to_H[i]: order[i] + 1 for i in range(len(order))}  # 1-index
        return smiles, order

    # region Ring Construction
    def _is_ring_aromatic(self, ring_atom_indices: list[(int, int)]):
        """
        Determines whether a ring is aromatic.

        Parameters
        ----------
        ring_atom_indices: the atom indices in the ring.

        Returns
        -------
        A boolean indicating whether this ring is aromatic.
        """
        for atom_idx1, atom_idx2 in self.get_bonds_in_ring(ring_atom_indices):
            bond = self.mol.GetBondBetweenAtoms(atom_idx1, atom_idx2)
            if not bond.GetIsAromatic():
                return False
        return True

    @staticmethod
    def _construct_old_graph(atom_list: list[Atom]):
        """
        To support older implementations of helper functions in Meeko, takes a list of atoms and uses it to create a
        list of each atom's graph value, where the index of a graph in the list corresponds to the atom's atom_index.

        Parameters
        ----------
        atom_list: list[Atom]
            A list of populated Atom objects.

        Returns
        -------
        A dict mapping from atom index to lists of ints, where each list of ints represents the bonds from that
        atom index to other atom indices.
        """
        output_graph = {}
        for atom in atom_list:
            output_graph[atom.index] = atom.graph
        return output_graph

    def perceive_rings(self, keep_chorded_rings: bool, keep_equivalent_rings: bool):
        """
        Uses Hanser-Jauffret-Kaufmann exhaustive ring detection to find the rings in the molecule

        Parameters
        ----------
        keep_chorded_rings: bool
            Indicates whether we want to keep chorded rings
        keep_equivalent_rings: bool
            Indicates whether we want to keep equivalent rings

        Returns
        -------
        None
        """
        old_graph = self._construct_old_graph(self.atoms)
        hjk_ring_detection = utils.HJKRingDetection(old_graph)
        rings = hjk_ring_detection.scan(keep_chorded_rings, keep_equivalent_rings)
        for ring_atom_indices in rings:
            ring_to_add = Ring(ring_atom_indices)
            if self._is_ring_aromatic(ring_atom_indices):
                ring_to_add.is_aromatic = True
            for atom_idx in ring_atom_indices:
                # TODO: add it to some sort of atom to ring id tracking thing -> add to atom data structure
                ring_to_add.graph = self._recursive_graph_walk(
                    atom_idx, collected=[], exclude=list(ring_atom_indices)
                )
            self.rings[ring_atom_indices] = ring_to_add
        return

    # endregion

    def get_conformer_with_modified_positions(self, new_atom_positions):
        """
        Gets a conformer with the specified new atom positions.
        We operate on one conformer at a time because SetTerminalAtomPositions acts on all conformers of a molecule,
        and we do not want to guarantee that all conformers require the same set of terminal atoms to be updated.

        Parameters
        ----------
        new_atom_positions:
            The new atom positions we want to use.

        Returns
        -------
        new_conformer:
            A new conformer with the input new atom positions.
        """
        new_mol = Chem.Mol(self.mol)
        new_conformer = Chem.Conformer(self.mol.GetConformer())
        is_set_list = [False] * self.mol.GetNumAtoms()
        for atom_index, new_position in new_atom_positions.items():
            new_conformer.SetAtomPosition(atom_index, new_position)
            is_set_list[atom_index] = True
        new_mol.RemoveAllConformers()
        new_mol.AddConformer(new_conformer, assignId=True)
        for atom_index, is_set in enumerate(is_set_list):
            if not is_set and new_mol.GetAtomWithIdx(atom_index).GetAtomicNum() == 1:
                neighbors = new_mol.GetAtomWithIdx(atom_index).GetNeighbors()
                if len(neighbors) != 1:
                    raise RuntimeError("Expected H to have one neighbors")
                Chem.SetTerminalAtomCoords(new_mol, atom_index, neighbors[0].GetIdx())
        return new_conformer

    def get_mol_with_modified_positions(self, new_atom_positions_list=None):
        """
        Modifies the stored RDKit Mol to a new set of atom positions, either those provided or the ones stored in
        self.modified_atom_positions, and returns the modified Mol object.

        Parameters
        ----------
        new_atom_positions_list:
            New atom positions to add to the RDKit Mol object.

        Returns
        -------
        new_mol: rdkit.Chem.rdchem.Mol
            A new RDKit Mol object with conformers that have the desired new atom positions.
        """
        if new_atom_positions_list is None:
            new_atom_positions_list = self.modified_atom_positions
        new_mol = Chem.Mol(self.mol)
        new_mol.RemoveAllConformers()
        for new_atom_positions in new_atom_positions_list:
            conformer = self.get_conformer_with_modified_positions(new_atom_positions)
            new_mol.AddConformer(conformer, assignId=True)
        return new_mol

    def get_num_mol_atoms(self):
        """
        Gets the number of atoms in the RDKit Mol object.

        Returns
        -------
        Number of atoms in the RDKit Mol object.
        """
        return self.mol.GetNumAtoms()

    def get_equivalent_atoms(self):
        """

        Returns
        -------

        """
        return list(Chem.CanonicalRankAtoms(self.mol, breakTies=False))

    @staticmethod
    def get_symmetries_for_rmsd(mol, max_matches=17):
        mol_noHs = Chem.RemoveHs(mol)
        matches = mol.GetSubstructMatches(
            mol_noHs, uniquify=False, maxMatches=max_matches
        )
        if len(matches) == max_matches:
            if mol.HasProp("_Name"):
                molname = mol.GetProp("_Name")
            else:
                molname = ""
            print(
                "warning: found the maximum nr of matches (%d) in RDKitMolSetup.get_symmetries_for_rmsd"
                % max_matches
            )
            print(
                'Maybe this molecule is "too" symmetric? %s' % molname,
                Chem.MolToSmiles(mol_noHs),
            )
        return matches

    @staticmethod
    def has_implicit_hydrogens(mol):
        # based on needsHs from RDKit's AddHs.cpp
        for atom in mol.GetAtoms():
            nr_H_neighbors = 0
            for neighbor in atom.GetNeighbors():
                nr_H_neighbors += int(neighbor.GetAtomicNum() == 1)
            if atom.GetTotalNumHs(includeNeighbors=False) > nr_H_neighbors:
                return True
        return False

    def restrain_to(
        self, target_mol, kcal_per_angstrom_square=1.0, delay_angstroms=2.0
    ):
        """

        Parameters
        ----------
        target_mol
        kcal_per_angstrom_square
        delay_angstroms

        Returns
        -------

        """
        if not _has_misctools:
            raise ImportError(_import_misctools_error)
        stereo_isomorphism = StereoIsomorphism()
        mapping, idx = stereo_isomorphism(target_mol, self.mol)
        lig_to_drive = {b: a for (a, b) in mapping}
        num_real_atoms = target_mol.GetNumAtoms()
        target_positions = target_mol.GetConformer().GetPositions()
        for atom_index in range(len(mapping)):
            target_xyz = target_positions[lig_to_drive[atom_index]]
            restraint = Restraint(
                atom_index, target_xyz, kcal_per_angstrom_square, delay_angstroms
            )
            self.restraints.append(restraint)
        return

    @staticmethod
    def from_json(obj):
        """
        Takes an object and attempts to decode it into an RDKitMoleculeSetup object.

        Parameters
        ----------
        obj: Object
            This can be any object, but it should be a dictionary generated by deserializing a JSON of an
            RDKitMoleculeSetup object.

        Returns
        -------
        If the input is a dictionary corresponding to a RDKitMoleculeSetup, will return an RDKitMoleculeSetup with data
        populated from the dictionary. Otherwise, returns the input object.
        """
        # if the input object is not a dict, we know that it will not be parsable and is unlikely to be usable or
        # safe data, so we should ignore it.
        if type(obj) is not dict:
            return obj

        # checks that all the keys we expect are in the object dictionary as a safety measure
        expected_molsetup_keys = {
            "mol",
            "modified_atom_positions",
            "dihedral_interactions",
            "dihedral_partaking_atoms",
            "dihedral_labels",
            "atom_to_ring_id",
            "ring_corners",
            "rmsd_symmetry_indices",
        }
        for key in expected_molsetup_keys:
            if key not in obj.keys():
                return obj

        # Uses the MoleculeSetup json decoder to get a base MoleculeSetup:
        base_molsetup = MoleculeSetup.from_json(obj)
        # Constructs an RDKitMoleculeSetup object
        rdkit_molsetup = RDKitMoleculeSetup()
        # Restores attributes from the MoleculeSetup. There may be a more elegant way to do this
        for key, value in base_molsetup.__dict__.items():
            rdkit_molsetup.__dict__[key] = deepcopy(value)
        # Restores RDKitMoleculeSetup-specific attributes from the json dict
        rdkit_molsetup.mol = rdkit_mol_from_json(obj["mol"])
        rdkit_molsetup.modified_atom_positions = [
            int(x) for x in obj["modified_atom_positions"]
        ]
        # TODO: Dihedral decoding may need another look
        rdkit_molsetup.dihedral_interactions = obj["dihedral_interactions"]
        rdkit_molsetup.dihedral_partaking_atoms = obj["dihedral_partaking_atoms"]
        rdkit_molsetup.dihedral_labels = obj["dihedral_labels"]
        rdkit_molsetup.atom_to_ring_id = {
            int(k): [string_to_tuple(t) for t in v]
            for k, v in obj["atom_to_ring_id"].items()
        }
        rdkit_molsetup.ring_corners = obj["ring_corners"]
        rdkit_molsetup.rmsd_symmetry_indices = [
            string_to_tuple(v) for v in obj["rmsd_symmetry_indices"]
        ]
        return rdkit_molsetup


# endregion


# region JSON Serialization and Deserialization
class AtomEncoder(json.JSONEncoder):
    """
    JSON Encoder class for Atom dataclass.
    """

    def default(self, obj):
        """
        Overrides the default JSON encoder for data structures for Atom objects.

        Parameters
        ----------
        obj: object
            Can take any object as input, but will only create the Atom JSON format for Atom objects.
            For all other objects will return the default JSON encoding.

        Returns
        -------
        A JSON serializable object that represents the Atom class or the default JSONEncoder output for an
        object.
        """
        if isinstance(obj, Atom):
            return {
                "index": obj.index,
                "pdbinfo": obj.pdbinfo,
                "charge": obj.charge,
                "coord": obj.coord.tolist(),  # converts coord from numpy array to lists
                "atomic_num": obj.atomic_num,
                "atom_type": obj.atom_type,
                "is_ignore": obj.is_ignore,
                "graph": obj.graph,
                "interaction_vectors": [v.tolist() for v in obj.interaction_vectors],
                "is_dummy": obj.is_dummy,
                "is_pseudo_atom": obj.is_pseudo_atom,
            }
        return json.JSONEncoder.default(self, obj)


class BondEncoder(json.JSONEncoder):
    """
    JSON Encoder class for Bond dataclass.
    """

    def default(self, obj):
        """
        Overrides the default JSON encoder for data structures for Bond objects.

        Parameters
        ----------
        obj: object
            Can take any object as input, but will only create the Bond JSON format for Bond objects.
            For all other objects will return the default JSON encoding.

        Returns
        -------
        A JSON serializable object that represents the Bond class or the default JSONEncoder output for an
        object.
        """
        if isinstance(obj, Bond):
            return {
                "canon_id": tuple_to_string(obj.canon_id),
                "index1": obj.index1,
                "index2": obj.index2,
                "rotatable": obj.rotatable,
            }
        return json.JSONEncoder.default(self, obj)


class RingEncoder(json.JSONEncoder):
    """
    JSON Encoder class for Ring dataclass.
    """

    def default(self, obj):
        """
        Overrides the default JSON encoder for data structures for Ring objects.

        Parameters
        ----------
        obj: object
            Can take any object as input, but will only create the Ring JSON format for Ring objects.
            For all other objects will return the default JSON encoding.

        Returns
        -------
        A JSON serializable object that represents the Ring class or the default JSONEncoder output for an
        object.
        """
        if isinstance(obj, Ring):
            return {
                "ring_id": tuple_to_string(obj.ring_id),
                "corner_flip": obj.corner_flip,
                "graph": obj.graph,
                "is_aromatic": obj.is_aromatic,
            }
        return json.JSONEncoder.default(self, obj)


class RestraintEncoder(json.JSONEncoder):
    """
    JSON Encoder class for Restraint dataclass.
    """

    def default(self, obj):
        """
        Overrides the default JSON encoder for data structures for Restraint objects.

        Parameters
        ----------
        obj: object
            Can take any object as input, but will only create the Restraint JSON format for Restraint objects.
            For all other objects will return the default JSON encoding.

        Returns
        -------
        A JSON serializable object that represents the Restraint class or the default JSONEncoder output for an
        object.
        """
        if isinstance(obj, Restraint):
            return {
                "atom_index": obj.atom_index,
                "target_coords": tuple_to_string(obj.target_coords),
                "kcal_per_angstrom_square": obj.kcal_per_angstrom_square,
                "delay_angstroms": obj.delay_angstroms,
            }
        return json.JSONEncoder.default(self, obj)


class MoleculeSetupEncoder(json.JSONEncoder):
    """
    JSON Encoder class for MoleculeSetup objects.
    """

    atom_encoder = AtomEncoder()
    bond_encoder = BondEncoder()
    ring_encoder = RingEncoder()
    restraint_encoder = RestraintEncoder()

    def default(self, obj):
        """
        Overrides the default JSON encoder for data structures for MoleculeSetup objects.

        Parameters
        ----------
        obj: object
            Can take any object as input, but will only create the MoleculeSetup JSON format for MoleculeSetup objects.
            For all other objects will return the default JSON encoding.

        Returns
        -------
        A JSON serializable object that represents the MoleculeSetup class or the default JSONEncoder output for an
        object.
        """
        output_dict = {}
        if isinstance(obj, MoleculeSetup):
            output_dict = {
                "name": obj.name,
                "is_sidechain": obj.is_sidechain,
                "pseudoatom_count": obj.pseudoatom_count,
                "atoms": [self.atom_encoder.default(x) for x in obj.atoms],
                "bond_info": {
                    tuple_to_string(k): self.bond_encoder.default(v)
                    for k, v in obj.bond_info.items()
                },
                "rings": {
                    tuple_to_string(k): self.ring_encoder.default(v)
                    for k, v in obj.rings.items()
                },
                "ring_closure_info": obj.ring_closure_info.__dict__,
                "rotamers": obj.rotamers,
                "atom_params": obj.atom_params,
                "restraints": [
                    self.restraint_encoder.default(x) for x in obj.restraints
                ],
                "flexibility_model": obj.flexibility_model,
            }
            # Addressing some flexibility model-specific structures.
            if "rigid_body_connectivity" in obj.flexibility_model:
                new_rigid_body_conn_dict = {
                    tuple_to_string(k): v
                    for k, v in obj.flexibility_model["rigid_body_connectivity"].items()
                }
                output_dict["flexibility_model"] = {
                    k: (
                        v
                        if k != "rigid_body_connectivity"
                        else new_rigid_body_conn_dict
                    )
                    for k, v in obj.flexibility_model.items()
                }
        # If it's an RDKitMoleculeSetup, adds the RDKitMoleculeSetup attributes
        if isinstance(obj, RDKitMoleculeSetup):
            output_dict["mol"] = rdMolInterchange.MolToJSON(obj.mol)
            output_dict["modified_atom_positions"] = obj.modified_atom_positions
            output_dict["dihedral_interactions"] = obj.dihedral_interactions
            output_dict["dihedral_partaking_atoms"] = obj.dihedral_partaking_atoms
            output_dict["dihedral_labels"] = obj.dihedral_labels
            output_dict["atom_to_ring_id"] = obj.atom_to_ring_id
            output_dict["ring_corners"] = obj.ring_corners
            output_dict["rmsd_symmetry_indices"] = obj.rmsd_symmetry_indices
        # If nothing is in the dict, then none of the possible object types for this encoder matched and we should
        # return the default JSON encoder.
        if len(output_dict) == 0:
            return json.JSONEncoder.default(self, obj)
        else:
            return output_dict


# endregion



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Meeko flexibility typer
#

from copy import deepcopy
from .utils import pdbutils

from .molsetup import Bond


def _calc_max_weighted_depth(
    model: dict,
    seed_node: int,
    bonds_to_break: tuple[tuple],
    visited: list[int] = None,
    depth: int = 0,
) -> int:
    """
    Scores a flexibility model based on the depth of the nesting. The number of atoms in a rigid group increases the
    weight of the bonds.

    Parameters
    ----------
    model: dict
        The flexibility model being moved over.
    seed_node: int
        Starting node index.
    bonds_to_break: tuple[tuple]
        An immutable list of bonds to break.
    visited: list[int]
        Nodes that have been visited.
    depth: int

    Returns
    -------
    max_value: int
        The flexibility model score.
    """
    glue_atoms = []
    # Adds the indices from bonds to break to the list of glue atoms
    for i, j in bonds_to_break:
        glue_atoms.append(i)
        glue_atoms.append(j)
    # Pulls the graph and members from the given flexibility model
    graph = model["rigid_body_graph"]
    members = model["rigid_body_members"]
    if visited is None:
        visited = []
    nr_atoms = len(members[seed_node])
    compensation = -1  # First atom after rotatable bond doesn't move, so doesn't weigh
    # Atoms in breakable bonds count twice, because of added complexity
    compensation += int(sum([i in glue_atoms for i in members[seed_node]]))
    this_value = depth**2 * (nr_atoms + compensation)
    max_value = this_value
    visited.append(seed_node)
    # Recurse on nodes that have not been visited yet.
    for node in graph[seed_node]:
        if node not in visited:
            visited.append(node)
            new_value = _calc_max_weighted_depth(
                model, node, bonds_to_break, visited, depth + 1
            )
            max_value = max(max_value, new_value + this_value)
    return max_value


def merge_terminal_atoms(flex_model: dict, not_terminal_atoms: list[int] = ()) -> None:
    """
    Rotatable bonds that link to a rigid body group that contains one atom are removed because that one atom lies on the
    bond axis and rotating the bond does not result in any movement of the atom. The atom after the removed rotatable
    bond is merged with the rigid body group that is upstream of the removed bond.

    Parameters
    ----------
    flex_model: dict
        Flexibility model
    not_terminal_atoms: list
        A list of non-terminal atoms

    Returns
    -------
    None
    """
    # Get members and graph from the flexibility model
    members_dict = flex_model["rigid_body_members"]
    graph = flex_model["rigid_body_graph"]
    remove = {}
    # Loops over the members dict and adds members that can be deleted to the removal dict
    for group_index, members in members_dict.items():
        if (
            len(members) == 1
            and len(graph[group_index]) == 1
            and members[0] not in not_terminal_atoms
            and group_index != flex_model["root"]
        ):
            remove[group_index] = members[0]
    # Goes through the removal dict and removes its contents from the flexibility model
    for group_index, atom_index in remove.items():
        host_index = graph[group_index][0]
        flex_model["rigid_body_members"].pop(group_index)
        flex_model["rigid_body_members"][host_index].append(atom_index)
        flex_model["rigid_body_graph"].pop(group_index)
        flex_model["rigid_body_graph"][host_index].remove(group_index)
        flex_model["rigid_body_connectivity"].pop((host_index, group_index))
        flex_model["rigid_body_connectivity"].pop((group_index, host_index))
        flex_model["rigid_index_by_atom"][atom_index] = host_index
    # subtracts the number of members we are removing from the count
    flex_model["rigid_body_count"] -= len(remove)
    return


def get_flexibility_model(
    molsetup,
    root_atom_index: int = None,
    break_combo_data: dict = None,
):
    """
    Given a MoleculeSetup, creates a flexibility model for that MoleculeSetup. Breaks macrocyclic rings if bond break
    information is provided.

    Parameters
    ----------
    molsetup: RDKitMoleculeSetup
        The molecule setup to generate a flexibility model for.
    root_atom_index: int

    break_combo_data: dict
        Data about different bond break combinations from the macrocycle typer.

    Returns
    -------
    model: dict
        The best flexibility model that could be generated given this MoleculeSetup and break data.
    broken_bonds: list
        A list of the bonds broken corresponding to the returned flexibility model.

    Raises
    ------
    RuntimeError:
        If the flexibility model generated ignores atoms that are not explicitly marked as needing to be ignored in the
        MoleculeSetup.
    """
    # If no macrocyclic rings are to be broken, we simply build the flexibility model.
    if break_combo_data is None or len(break_combo_data["bond_break_combos"]) == 0:
        bonds_to_break = ()
        unbroken_rings_bonds = []
        # Gets all bonds from the MoleculeSetup
        for ring in molsetup.rings:
            for bond in molsetup.get_bonds_in_ring(ring):
                unbroken_rings_bonds.append(bond)
        # Instantiates the flexibility model
        flex_model = walk_rigid_body_graph(
            molsetup, bonds_to_break, unbroken_rings_bonds
        )
        # Gets the number of atoms that aren't marked to be ignored
        nr_not_ignored = sum(
            [not molsetup.get_is_ignore(i) for i in range(len(molsetup.atoms))]
        )
        # Checks the validity of the model by ensuring that all of the atoms that should have been visited were visited,
        # otherwise raises an error
        if len(flex_model["visited"]) != nr_not_ignored:
            molsetup.show()
            msg = f"{len(flex_model['visited'])=} differs from not-ignored atoms {nr_not_ignored}"
            raise RuntimeError(msg)
        # Sets the model root body index
        root_body_index = get_root_body_index(flex_model, root_atom_index)
        flex_model["root"] = root_body_index
        broken_bonds = []
        # Returns the model and the list of broken bonds
        return flex_model, broken_bonds

    # The macrocycle typer enumerated rings to break, and a number of lists
    # of bonds to break/delete that result in breaking of the enumerated rings.
    # to be deleted to break each ring. Deleting a bond results in the remaining
    # bonds within the same ring to become rotatable.
    bond_break_combos = break_combo_data["bond_break_combos"]
    bond_break_scores = break_combo_data["bond_break_scores"]
    unbroken_rings_list = break_combo_data["unbroken_rings"]
    best_model = None
    best_score = float("+inf")
    best_index = None
    # Loops through the potential bond break combinations, generates flexibility model information for each one, and
    # compares the flexibility models to determine the best model to return.
    for index in range(len(bond_break_combos)):
        bond_break_combo = bond_break_combos[index]
        bond_break_score = bond_break_scores[index]
        unbroken_rings_bonds = []
        for ring in unbroken_rings_list[index]:
            for bond in molsetup.get_bonds_in_ring(ring):
                unbroken_rings_bonds.append(bond)

        # Generates a model
        flex_model = walk_rigid_body_graph(
            molsetup, bond_break_combo, unbroken_rings_bonds
        )

        # Gets the number of atoms that aren't marked to be ignored
        nr_not_ignored = sum(
            [not molsetup.get_is_ignore(i) for i in range(len(molsetup.atoms))]
        )
        # Checks the validity of the model by ensuring that all of the atoms that should have been visited were visited,
        # otherwise raises an error
        if len(flex_model["visited"]) != nr_not_ignored:
            msg = f"{len(flex_model['visited'])=} differs from not-ignored atoms {nr_not_ignored}"
            raise RuntimeError(msg)
        # Sets the model root body index
        root_body_index = get_root_body_index(flex_model, root_atom_index)
        flex_model["root"] = root_body_index
        depth_weighted = _calc_max_weighted_depth(
            flex_model, flex_model["root"], bond_break_combo
        )
        # larger bond_break_score is better, larger depth is worse
        # bond break score kinda disappeared in another branch (bonds are either breakable or not)
        score = depth_weighted - 0.001 * bond_break_score
        if score < best_score:
            best_score = score
            best_model = flex_model
            best_index = index

    best_model["score"] = best_score
    broken_bonds = list(break_combo_data["bond_break_combos"][best_index])
    # Returns the best model and the list of broken bonds
    return best_model, broken_bonds


def get_root_body_index(model: dict, root_atom_index: int = None) -> int:
    """
    Gets the index of the rigid body group in the flexibility model that contains the given root_atom_index

    Parameters
    ----------
    model: dict
        Flexibility model
    root_atom_index: int


    Returns
    -------
    root_body_index: int
    """

    # find and return index of rigid body group that contains root_atom_index
    if root_atom_index is not None:
        for body_index in model["rigid_body_members"]:
            if root_atom_index in model["rigid_body_members"][body_index]:
                # 1-index atoms
                return body_index

    # find rigid group that minimizes weighted graph depth
    graph = deepcopy(model["rigid_body_graph"])
    while len(graph) > 2:  # remove leaves until 1 or 2 rigid groups remain
        leaves = []
        for vertex, edges in list(graph.items()):
            if len(edges) == 1:
                leaves.append(vertex)
        for leaf in leaves:
            for vertex, edges in list(graph.items()):
                if leaf in edges:
                    edges.remove(leaf)
                    graph[vertex] = edges
            del graph[leaf]
    if len(graph) == 0:
        root_body_index = 0
    elif len(graph) == 1:
        root_body_index = list(graph.keys())[0]
    else:
        r1, r2 = list(graph.keys())
        r1_size = len(model["rigid_body_members"][r1])
        r2_size = len(model["rigid_body_members"][r2])
        if r1_size >= r2_size:
            root_body_index = r1
        else:
            root_body_index = r2
    return root_body_index


def update_closure_atoms(
    molsetup, bonds_to_break: list[tuple], glue_pseudo_atoms: dict
) -> None:
    """
    Create pseudoatoms required by breaking bonds in the flexibility model

    Parameters
    ----------
    molsetup: MoleculeSetup
        Molecule Setup to modify and add pseudoatoms to
    bonds_to_break: tuple
        List of bonds to break
    glue_pseudo_atoms: dict

    Returns
    -------
    None
    """
    # Loops through the bonds to break and adds all the bonds to
    for i, bond in enumerate(bonds_to_break):
        molsetup.ring_closure_info.bonds_removed.append(
            bond
        )  # bond is a pair of atom indices

        # calculate position and parameters of the pseudoatoms for the closure
        for idx in (0, 1):
            target = bond[1 - idx]
            anchor = bond[0 - idx]
            if glue_pseudo_atoms is None or len(glue_pseudo_atoms) == 0:
                coord = molsetup.get_coord(target)
            else:
                coord = glue_pseudo_atoms[anchor]
            anchor_info = molsetup.get_pdbinfo(anchor)
            pdbinfo = pdbutils.PDBAtomInfo(
                "G",
                anchor_info.resName,
                anchor_info.resNum,
                anchor_info.icode,
                anchor_info.chain,
            )
            pseudo_index = molsetup.add_pseudoatom(
                coord=coord,
                charge=0.0,
                anchor_list=[anchor],
                atom_type=f"G{i}",
                rotatable=False,
                pdbinfo=pdbinfo,
            )
            if anchor in molsetup.ring_closure_info.pseudos_by_atom:
                raise RuntimeError("did not expect more than one G per atom")
            molsetup.ring_closure_info.pseudos_by_atom[anchor] = pseudo_index
        molsetup.set_atom_type(bond[0], "CG%d" % i)
        molsetup.set_atom_type(bond[1], "CG%d" % i)
    return


def walk_rigid_body_graph(
    molsetup,
    bonds_to_break: tuple,
    unbroken_rings_bonds: list[tuple],
    start: int = None,
    data: dict = None,
):
    """
    Recursively walks through the MoleculeSetup to build a graph of rigid bodies. Uses that graph to create and
    populate a flexibility model.

    Parameters
    ----------
    molsetup: RDKitMoleculeSetup
        MoleculeSetup to walk through
    bonds_to_break: tuple[tuple]
    unbroken_rings_bonds: list[tuple(int, int]
    start: int
    data: dict

    Returns
    -------
    data: dict
        A dictionary representing a flexibility model
    """
    # If start is none, uses the default start which is the first non-ignored atom
    if start is None:
        for atom in molsetup.atoms:
            if not atom.is_ignore:
                start = atom.index
                break
    # If the start atom is marked to be ignored, returns nothing
    if molsetup.get_is_ignore(start):
        return
    if data is None:
        data = {
            "visited": [],
            "rigid_body_count": 0,
            "rigid_index_by_atom": {},
            "rigid_body_members": {},
            "rigid_body_connectivity": {},
            "rigid_body_graph": {},
        }
    data["visited"].append(start)
    data["rigid_index_by_atom"][start] = data["rigid_body_count"]
    rigid_index = data["rigid_body_count"]
    sprouts_buffer = []
    # The while loop goes on until all atoms in `group_members` are queried
    # for their neighbors. `idx` is the index within `group_members`, and
    # `current` is the index of the atom being queried. When a neighbor is
    # bonded to `current` by a non-rotatable/rigid bond, it is added to
    # `group_members`. Remember that `idx` is not an index of an atom.
    idx = 0
    group_members = [start]
    while idx < len(group_members):
        current = group_members[idx]
        if molsetup.get_is_ignore(current):
            idx += 1
            continue
        for neigh in molsetup.get_neighbors(current):
            if molsetup.get_is_ignore(neigh):
                continue
            bond_id = Bond.get_bond_id(current, neigh)
            if bond_id in bonds_to_break:
                continue
            bond_info = None
            if bond_id in molsetup.bond_info:
                bond_info = molsetup.bond_info[bond_id]
            if neigh in data["visited"]:
                neigh_in_other_rigid_body = (
                    rigid_index != data["rigid_index_by_atom"][neigh]
                )
                if not bond_info.rotatable and neigh_in_other_rigid_body:
                    raise RuntimeError(
                        "Flexible bonds within rigid group. We have a problem."
                    )
                continue
            if bond_info.rotatable and bond_id not in unbroken_rings_bonds:
                sprouts_buffer.append((current, neigh))
            else:
                group_members.append(neigh)
                data["rigid_index_by_atom"][neigh] = rigid_index
                data["visited"].append(neigh)
        idx += 1
    data["rigid_body_members"][rigid_index] = list(group_members)
    data["rigid_body_graph"].setdefault(rigid_index, [])
    for i, (current, neigh) in enumerate(sprouts_buffer):
        if neigh in data["visited"]:
            continue
        data["rigid_body_count"] += 1
        next_rigid_index = data["rigid_body_count"]
        data["rigid_body_connectivity"][rigid_index, next_rigid_index] = current, neigh
        data["rigid_body_connectivity"][next_rigid_index, rigid_index] = neigh, current
        data["rigid_body_graph"].setdefault(next_rigid_index, [])
        data["rigid_body_graph"][rigid_index].append(next_rigid_index)
        data["rigid_body_graph"][next_rigid_index].append(rigid_index)
        walk_rigid_body_graph(
            molsetup, bonds_to_break, unbroken_rings_bonds, neigh, data
        )
    return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#

import argparse
import gzip
import pathlib
import sys
from os import linesep as eol
import warnings
import copy
import numpy as np

from rdkit import Chem

from meeko import PDBQTMolecule
from meeko import RDKitMolCreate
from meeko import Polymer
from meeko import export_pdb_updated_flexres
from meeko.utils.utils import parse_begin_res


def cmd_lineparser():
    parser = argparse.ArgumentParser(
        description='Export docked ligand to SDF, and receptor to PDB',
    )
    parser.add_argument(dest='docking_results_filename', nargs = "+",
                        help='Docking output file(s), either PDBQT \
                        file from Vina or DLG file from AD-GPU.')
    parser.add_argument('-s', '--write_sdf', metavar='output_SDF_filename',
                        help="defaults to input filename with suffix from --sufix")
    parser.add_argument(
        '-p',
        '--write_pdb',
        metavar='output_PDB_filename',
        help="defaults to input filename with suffix from --suffix",
    )
    parser.add_argument("--suffix", default="_docked",
                        help="suffix for output filesnames that are not explicitly set")
    parser.add_argument('-j', '--read_json', metavar='input_JSON_filename',
                        help="receptor written by mk_prepare_receptor -j/--write_json")
    parser.add_argument('--all_dlg_poses', action='store_true',
                        help="write all AutoDock-GPU poses, not just cluster leads.")
    parser.add_argument('-k', '--keep_flexres_sdf', action='store_true',
                        help="add flexres, if any, to SDF ouput")
    parser.add_argument('-', '--',  dest='redirect_stdout', action='store_true',
                        help="do not write SDF file, just print it to STDOUT")
    return parser.parse_args()

def main():
    args = cmd_lineparser()
    
    docking_results_filenames = args.docking_results_filename
    write_sdf = args.write_sdf
    write_pdb = args.write_pdb
    read_json = args.read_json
    suffix = args.suffix
    all_dlg_poses = args.all_dlg_poses
    keep_flexres_sdf = args.keep_flexres_sdf
    redirect_stdout = args.redirect_stdout
    
    if (
        (write_sdf is not None or write_pdb is not None) and
        len(docking_results_filenames) > 1
    ):
        msg = "With multiple input files, the output filenames are based on the"
        msg += "input filename. The suffix can be controlled with option --suffix."
        msg += "--write options are incompatible with multiple input files"
        print("--write options incompatible with multiple input files", file=sys.stderr)
        sys.exit(2)
    
    if redirect_stdout and len(docking_results_filenames) > 1:
        print("option -/-- incompatible with multiple input files", file=sys.stderr)
        sys.exit(2)
    
    if read_json is not None:
        with open(read_json) as f:
            json_string = f.read()
        polymer = Polymer.from_json(json_string)
    else:
        polymer = None
        if write_pdb is not None:
            print("option -p (write pdb) requires -j (receptor receptor file)", file=sys.stderr)
            sys.exit(2)
    
    
    for filename in docking_results_filenames:
        is_dlg = filename.endswith('.dlg') or filename.endswith(".dlg.gz")
        if filename.endswith(".gz"):
            with gzip.open(filename) as f:
                string = f.read().decode()
        else:
            with open(filename) as f:
                string = f.read()
        mol_name = pathlib.Path(filename).with_suffix("").name
        pdbqt_mol = PDBQTMolecule(
            string,
            name=mol_name,
            is_dlg=is_dlg,
            skip_typing=True,
        )
        only_cluster_leads = is_dlg and not all_dlg_poses
        sdf_string, failures = RDKitMolCreate.write_sd_string(
                pdbqt_mol,
                only_cluster_leads=only_cluster_leads,
                keep_flexres=keep_flexres_sdf,
        )
        for i in failures:
            warnings.warn("molecule %d not converted to RDKit/SD File" % i)
        if sdf_string == "": 
            warnings.warn("sdf_string does not contain molecular data.")
            if redirect_stdout or write_pdb: 
                pass
            else:
                print("Output SDF will not be created because there is no pose data for ligand. \n"
                        + "Maybe the input poses only contain flexible sidechains and \n"
                        + "keep_flexres_sdf is set to False. \n"
                        + "Use -k with mk_export.py to retain the flexres and write to output SDF File. ")
                sys.exit(2)
        if len(failures) == len(pdbqt_mol._atom_annotations["mol_index"]):
            msg = "\nCould not convert to RDKit. Maybe meeko was not used for preparing\n"
            msg += "the input PDBQT for docking, and the SMILES string is missing?\n"
            msg += "Except for standard protein sidechains, all ligands and flexible residues\n"
            msg += "require a REMARK SMILES line in the PDBQT, which is added automatically by meeko."
            raise RuntimeError(msg)
        if not redirect_stdout:
            if write_sdf is None:
                fn = pathlib.Path(filename).with_suffix("").name + f"{suffix}.sdf"
            else:
                fn = write_sdf
            with open(fn, "w") as f:
                f.write(sdf_string)
        else:
            print(sdf_string)
    
        # write receptor with updated flexres
        if read_json is not None:
            pdb_string = ""
            pose_id_to_iter = [pose.pose_id for pose in pdbqt_mol]
            iter_pose = copy.deepcopy(pdbqt_mol)
            for pose_id in pose_id_to_iter:
                model_nr = pose_id + 1
                iter_pose._positions = np.array([pdbqt_mol._positions[pose_id]])
                iter_pose._pose_data['n_poses'] = 1  # Update the number of poses to reflect the reduction
                iter_pose._current_pose = 0  # Reset to the first (and only) pose
                pdb_string += "MODEL " + f"{model_nr:8}" + eol
                pol_copy = copy.deepcopy(polymer)
                pdb_string += export_pdb_updated_flexres(pol_copy, iter_pose)
                pdb_string += "ENDMDL" + eol
            if write_pdb is None:
                fn = pathlib.Path(filename).with_suffix("").name + f"{suffix}.pdb"
            else:
                fn = write_pdb
            with open(fn, "w") as f:
                f.write(pdb_string)
    return

if __name__ == "__main__":
    sys.exit(main())



from typing import Union
from typing import Optional
import numpy as np
from rdkit import Chem
from rdkit.Chem import rdDetermineBonds
from rdkit.Chem.rdchem import Mol
from .rdkitutils import AtomField
from .rdkitutils import build_one_rdkit_mol_per_altloc
from .rdkitutils import _aux_altloc_mol_build

from prody.atomic import ATOMIC_FIELDS
from prody.atomic.atomgroup import AtomGroup
from prody.atomic.selection import Selection
from prody.atomic.segment import Segment
from prody.atomic.chain import Chain
from prody.atomic.residue import Residue

# used to convert from and to element symbols ans atomic number
periodic_table = Chem.GetPeriodicTable()

# The following are not used yet...
# https://chemicbook.com/2021/03/01/how-to-show-atom-numbers-in-rdkit-molecule.html
_bondtypes = {
    1: Chem.BondType.SINGLE,
    2: Chem.BondType.DOUBLE,
    3: Chem.BondType.TRIPLE,
    6: Chem.BondType.AROMATIC,
}
_bondstereo = {
    0: Chem.rdchem.BondStereo.STEREONONE,
    1: Chem.rdchem.BondStereo.STEREOE,
    2: Chem.rdchem.BondStereo.STEREOZ,
}
prody_atom_properties = {"sec_structure": "getSecstr"}
names = Chem.rdchem.AtomMonomerType.names
###


# TODO WORK IN PROGRESS
# NOT WORKING
def rdkit_to_prody(
    rdkit_mol: Mol,
    name: str = "",
    conformer_idx: int = 0,
    default_chain: str = "R",
    default_resname: str = "UNK",
):
    """
    THIS FUNCTION IS INCOMPLETE
    """
    if name == "":
        name = rdkit_mol.GetProp("_Name")
    if name == "" or name is None:
        name = "rdkit_to_prody_molecule"
    prody_obj = AtomGroup(name)
    idx_prody_to_rdkit = {}
    idx_rdkit_to_prody = {}
    idx_rdkit = 0
    atom_count = rdkit_mol.GetNumAtoms()
    # get coordinates
    conformer = rdkit_mol.GetConformer(conformer_idx)
    coord_array = conformer.GetPositions()

    # see python4.9/site-packages/prody/compounds/pdbligands.py]
    chain_array = np.zeros(atom_count, dtype=ATOMIC_FIELDS["chain"].dtype)
    resname_array = np.zeros(atom_count, dtype=ATOMIC_FIELDS["resname"].dtype)
    resnum_array = np.zeros(atom_count, dtype=ATOMIC_FIELDS["resnum"].dtype)
    aname_array = np.zeros(atom_count, dtype=ATOMIC_FIELDS["name"].dtype)
    atype_array = np.zeros(atom_count, dtype=ATOMIC_FIELDS["type"].dtype)
    # properties
    hetero_array = np.zeros(atom_count, dtype=bool)
    serial_array = np.zeros(atom_count, dtype=ATOMIC_FIELDS["serial"].dtype)
    segment_array = np.zeros(atom_count, dtype=ATOMIC_FIELDS["segment"].dtype)
    element_array = np.zeros(atom_count, dtype=ATOMIC_FIELDS["element"].dtype)

    # iterate on atoms
    # see issue https://github.com/rdkit/rdkit/issues/6208
    # to be fixed in 2023.09 version of RDKit.
    for i in range(0, rdkit_mol.GetNumAtoms()):
        atom = rdkit_mol.GetAtomWithIdx(i)
        atomic_num = atom.GetAtomicNumber()
        element_array[i] = periodic_table.GetElement(atomic_num)
        if hasattr(rdkit_mol, "prody_to_rdkit"):
            serial_array[i] = rdkit_mol.prody_to_rdkit[i]
        else:
            serial_array[i] = i + 1

        # gather residue info
        chain_array
        resname_array
        resnum_array
        aname_array
        segment_array

        # if non_std residue, thebmn set flag hetero
        hetero_array

    # charges = np.zeros(asize, dtype=ATOMIC_FIELDS["charge"].dtype)
    # radii = np.zeros(asize, dtype=ATOMIC_FIELDS["radius"].dtype)

    for bond in rdkit_mol.GetBonds():
        pass

    # https://sourceforge.net/p/rdkit/mailman/rdkit-discuss/thread/60825b0f0804152055x4ba39e13k7d62f81634413ca6@mail.gmail.com/


ALLOWED_PRODY_TYPES = Union[Selection, AtomGroup, Chain, Residue]


def prody_to_rdkit(
    prody_obj: ALLOWED_PRODY_TYPES,
    sanitize: bool = True,
    keep_bonds: bool = False,
    requested_altloc: Optional[str] = None,
    default_altloc: Optional[str] = None,
) -> Mol:
    """
    Convert a ProDy selection or atom group into an RDKit molecule.

    The function accepts both Selections or AtomGroups and generates the
    correspodning RDKit object, retaining most of the information available
    from the ProDy object into RDKit atoms using the AtomPDBResidueInfo class.

    By default, only coordinates, atom element and name,
    and residue info are retained. Bond and charge info,
    even if available are ignored.

    Two convenience lookup tables are attached to the new RDKit molecule to use
    ProDy indices to retrieve the corresponding RDKit indices
    (`mol._idx_prody_to_rdkit`) and vice versa (`mol._idx_rdkit_to_prody`), e.g.:

    >>> tyr_oh = prody_mol.select("resname TYR and name OH and resnum 108 and chid A")
    <Selection: 'resname TYR and... 108 and chid A' from 1jff (1 atoms)>
    >>> idx = tyr_oh.getIndices()[0]
    >>> rdkit_mol._idx_prody_to_rdkit[idx]
    11
    >>> original_res_atom_idx = rdkit_mol._idx_rdkit_to_prody[11]
    640
    >>> rdkit_mol._idx_rdkit_to_prody
    {0: 629, 1: 630, 2: 631, 3: 632, 4: 633, 5: 634, 6: 635, 7: 636, 8: 637, 9:
    638, 10: 6 39, 11: 640}

    By default, even if available, bonds defined in the ProDy object are
    ignored. In order to use ProDy bond info, the flag `keep_bonds` needs to
    be set to True and bonds assigned or perceived prior to the conversion
    (e.g. `AtomGroup.inferBonds()`, which is usually slow). As of v2.4.1,
    Prody's mmCIF parser does not parse bonds.

    Parameters
    ----------
    prody_obj : Union[Selection, AtomGroup]
        input molecule (AtomGroup) or subset selection (e.g., Selection
        containign one residue)
    sanitize : bool (default: True)
        perform the sanitization of the RDKit molecule prior to returning it;
        any exception during this process will be raised as RDKit would do.
    keep_bonds : bool (default: False)
        use bonds defined in the ProDy object, instead of perceiving them with
        RDKit using rdDetermineBonds.DetermineConnectivity()
        Note that prody's mmCif parser does not parse bonds as of v2.4.1.
    wanted_altloc : Optional[str] (default: None)
        require specified alternate location
    default_altloc : Optional[str] (default: None)
        if altlocs exist, use this one.

    Returns
    -------
    raw_input_mols : dict of tuples
        RDKit molecule
    needed_altloc : bool
        need to know which altloc to use with wanted_altloc or default_altloc
    missed_altloc : bool
        the requested altloc (with wanted_altloc) does not exist
    """

    atom_field_list = []
    indices_prody = []
    for atom in prody_obj:
        indices_prody.append(int(atom.getIndex()))
        x, y, z = atom.getCoords()
        atom_field = AtomField(
            atomname=str(atom.getName()).strip(),
            altloc=str(atom.getAltloc()).strip(),
            resname=str(atom.getResname()).strip(),
            chain=str(atom.getChid()).strip(),
            resnum=int(atom.getResnum()),
            icode=str(atom.getIcode()).strip(),
            x=float(x),
            y=float(y),
            z=float(z),
            element=str(atom.getElement()).strip(),
        )
        atom_field_list.append(atom_field)
    
    rdmol, idx_to_rdkit, missed_altloc, needed_altloc = _aux_altloc_mol_build(
        atom_field_list,
        requested_altloc,
        default_altloc,
    )

    if rdmol is None:
        return None, missed_altloc, needed_altloc

    idx_prody_to_rdkit = {}
    idx_rdkit_to_prody = {}
    for index_list, index_rdkit in idx_to_rdkit.items():
        index_prody = indices_prody[index_list]
        idx_prody_to_rdkit[index_prody] = index_rdkit
        idx_rdkit_to_prody[index_rdkit] = index_prody

    # iterate and create bonds
    if keep_bonds is True:
        if isinstance(prody_obj, AtomGroup):
            bonds = prody_obj.getBonds()
        else:
            bonds = prody_obj.getAtomGroup().getBonds()
    else:
        bonds = None
    if bonds is not None:
        print("THIS CODE PATH IS UNTESTED (bonds in prody_to_rdkit)")
        mol = Chem.EditableMol(rdmol)
        mol.BeginBatchEdit()
        for bond in bonds:
            bond_indices = [int(x) for x in bond.getIndices()]
            # bond with atoms outside this selection
            if (
                not bond_indices[0] in idx_rdkit_to_prody
                or not bond_indices[1] in idx_rdkit_to_prody
            ):
                continue
            # bond_type = bond_types[bond[-1]]
            bond_order = 1
            bond_type = _bondtypes[bond_order]
            print("bonds", bond_indices, bond)
            mol.AddBond(bond_indices[0], bond_indices[1], bond_type)

        mol.CommitBatchEdit()
        rdmol = mol.GetMol()

        # the molecule needs bonds, one way or another
    if bonds is None:
        rdDetermineBonds.DetermineConnectivity(rdmol)
    # sanitize the molecule if necessary
    # TODO before or after ond perception?
    if sanitize:
        Chem.SanitizeMol(rdmol)
    # attach the bookkeeping to the molecule
    rdmol._idx_prody_to_rdkit = idx_prody_to_rdkit
    rdmol._idx_rdkit_to_prody = idx_rdkit_to_prody
    return rdmol, missed_altloc, needed_altloc


if __name__ == "__main__":
    import sys
    import prody

    # # test rdkit
    # mol = Chem.SDMolSupplier("rdkit_input.sdf")[0]
    # prody_mol = rdkit_to_prody(mol)
    # sys.exit(0)
    # test prody
    # prot = prody.parseMMCIF("1crn", headers=True)
    pdb_id = "1jff"
    selection_string = "resname TA1"
    if len(sys.argv) > 1:
        pdb_id = sys.argv[1]
        selection_string = sys.argv[2]
    print("- parsing protein %s..." % pdb_id, end="")
    prot = prody.parseMMCIF(pdb_id, headers=True)

    # uncomment this to perceive bonds with ProDy (slow)
    # prot.inferBonds()
    print(">> %d atoms" % len(prot))
    # http://prody.csb.pitt.edu/manual/reference/atomic/select.html#module-prody.atomic.select
    sel = prot.select(selection_string)
    # sel = prot.select("resname THR resnum 1")
    print('- selected %d atoms with the string "%s"' % (len(sel), selection_string))
    # x = input()
    print("- converting selection to RDKit...")
    frag = prody_to_rdkit(sel)
    writer = Chem.SDWriter("fragment.sdf")
    writer.write(frag)
    writer.close()
    print("  [ written fragment.sdf ]")

    print("- converting the entire input as RDKit...")
    whole = prody_to_rdkit(prot)
    writer = Chem.SDWriter("whole.sdf")
    writer.write(whole)
    writer.close()
    print("  [ written whole.sdf ]")



#!/usr/bin/env python

import argparse
import json
import math
from os import linesep as eol
import pathlib
import sys

import numpy as np

from meeko.reactive import atom_name_to_molsetup_index, assign_reactive_types_by_index
from meeko import PDBQTMolecule
from meeko import RDKitMolCreate
from meeko import MoleculePreparation
from meeko import MoleculeSetup
from meeko import ResidueChemTemplates
from meeko import PDBQTWriterLegacy
from meeko import Polymer
from meeko import PolymerEncoder
from meeko import PolymerCreationError
from meeko import reactive_typer
from meeko import get_reactive_config
from meeko import gridbox
from meeko import __file__ as pkg_init_path
from rdkit import Chem

try:
    import prody
except ImportError as import_error:
    _prody_import_error = import_error
    _got_prody = False
else:
    SUPPORTED_PRODY_FORMATS = {"pdb": prody.parsePDB, "cif": prody.parseMMCIF}
    _got_prody = True

path_to_this_script = pathlib.Path(__file__).resolve()


def parse_cmdline_res(string):
    """ "A:5,7,BB:12C  ->  "A:5", "A:7", "BB:12C" """
    blocks = ("," + string).split(":")
    nr_blocks = len(blocks) - 1
    keys = []
    for i in range(nr_blocks):
        chain = blocks[i].split(",")[-1]
        if i + 1 == nr_blocks:
            resnums = blocks[i + 1].split(",")
        else:
            resnums = blocks[i + 1].split(",")[:-1]
        if len(resnums) == 0:
            raise ValueError(f"missing residue in {resnums}")
        for resnum in resnums:
            keys.append(f"{chain}:{resnum}")
    return keys


def parse_cmdline_res_assign(string):
    """convert "A:5,7=CYX,A:19A,B:17=HID" to {"A:5": "CYX", "A:7": "CYX", ":19A": "HID"}"""

    output = {}
    nr_assignments = string.count("=")
    string = "," + string  # enables `residues =` below to work in first iteraton
    tmp = string.split("=")
    for i in range(nr_assignments):
        residues = tmp[i].split(",")[1:]
        assigned_name = tmp[i + 1].split(",")[0]
        chain = ""
        for residue in residues:
            fields = residue.split(":")
            if len(fields) == 1:
                resnum = fields[0]
            elif len(fields) == 2:
                chain = fields[0]
                resnum = fields[1]
            else:
                raise ValueError(f"too many : in {residue}")
            if len(resnum) == 0:
                raise ValueError(f"missing residue in {residues}")
            key = f"{chain}:{resnum}"
            if key in output:
                raise ValueError(f"repeated {key} in {residue}")
            output[key] = assigned_name
    return output


class TalkativeParser(argparse.ArgumentParser):
    def error(self, message):
        """overload to print_help for every error"""
        self.print_help()
        this_script = path_to_this_script.name
        print("\n%s: error: %s" % (this_script, message), file=sys.stderr)
        sys.exit(2)


def check(success, error_msg):
    if not success:
        print("Error: " + error_msg, file=sys.stderr)
        sys.exit(2)

def required_length(nmin, nmax):
    class RequiredLength(argparse.Action):
        def __call__(self, parser, namespace, values, option_string=None):
            if not nmin <= len(values) <= nmax:
                msg = "fargument {self.dest} requires between"
                msg += " {nmin} and {nmax} arguments"
                raise argparse.ArgumentTypeError(msg)
            setattr(namespace, self.dest, values)
    return RequiredLength

def get_args():
    parser = TalkativeParser()

    io_group = parser.add_argument_group("Input/Output")
    io_group.add_argument(
        "--read_pdb",
        metavar="PDB_FILENAME",
        help="reads PDB, not PDBQT, and does not use ProDy",
    )
    need_prody_msg = ""
    # if prody is not installed, the help message is extended to tell
    # the user how to install prody
    if not _got_prody:
        need_prody_msg = " which can be installed from PyPI or conda-forge."
    io_group.add_argument(
        "-i",
        "--read_with_prody",
        metavar="MACROMOL_FILENAME",
        help=f"reads PDB/mmCIF file with Prody{need_prody_msg}")
    io_group.add_argument(
        "-o",
        "--output_basename",
        help="default basename for --write options used without filename",
    )
    io_group.add_argument(
        "-p", "--write_pdbqt",
        metavar="PDBQT_FILENAME",
        nargs="*",
        help="adds _rigid/_flex with flexible residues (filename defaults to --output_basename when not specified)",
    )
    io_group.add_argument(
        "-j", "--write_json",
        metavar="JSON_FILENAME",
        help="parameterized receptor (filename defaults to --output_basename when not specified)",
        nargs="*",
        action=required_length(0, 1))

    io_group.add_argument(
        "--write_pdb",
        help="prepared receptor (must specify filename)",
        nargs="*",
        metavar="PDB_FILENAME",
    )
    io_group.add_argument(
        "-g",
        "--write_gpf",
        metavar="GPF_FILENAME",
        help="autogrid input file (filename defaults to --output_basename when not specified)",
        nargs="*",
        action=required_length(0, 1))
    io_group.add_argument(
        "-v", "--write_vina_box",
        metavar="VINA_BOX_FILENAME",
        help="config file for Vina with box dimensions (filename defaults to --output_basename when not specified_",
        nargs="*",
        action=required_length(0, 1))


    config_group = parser.add_argument_group("Receptor perception")
    config_group.add_argument("-n", "--set_template", help="e.g. A:5,7=CYX,B:17=HID")
    config_group.add_argument("-d", "--delete_residues", help="e.g. A:350,B:15,16,17")
    config_group.add_argument("-b", "--blunt_ends", help="e.g. A:123,200=2,A:1=0")
    config_group.add_argument("--add_templates", help="[.json]", metavar="JSON_FILENAME", nargs="+", default=[])
    config_group.add_argument("--mk_config", help="[.json]", metavar="JSON_FILENAME")
    config_group.add_argument(
        "-a", "--allow_bad_res",
        action="store_true",
        help="delete residues with missing atoms instead of raising error",
    )
    config_group.add_argument("--default_altloc", help="default alternate location (overridden by --wanted_altloc)")
    config_group.add_argument("--wanted_altloc", help="require altloc for specific residues, e.g. :5=B,B:17=A")
    config_group.add_argument(
        "-f",
        "--flexres",
        action="append",
        default=[],
        help='specify the flexible residues by the chain ID and residue number, e.g. -f ":42,B:23" is equivalent to -f ":42" -f "B:23" (leave chain ID empty if omitted in input PDB or mmCIF)',
    )

    box_group = parser.add_argument_group("Size and center of grid box")

    box_group.add_argument(
        "--box_size", help="size of grid box (x, y, z) in Angstrom", nargs=3, type=float,
        metavar=("X", "Y", "Z"),
    )
    box_group.add_argument(
        "--box_center",
        help="center of grid box (x, y, z) in Angstrom",
        nargs=3,
        metavar=("X", "Y", "Z"),
        type=float,
    )
    box_group.add_argument(
        "--box_center_off_reactive_res",
        help="project grid box center along CA-CB bond 5 A away from CB (only applicable when there is exactly one reactive flexible residue)",
        action="store_true",
    )
    box_group.add_argument(
        "--box_enveloping",
        metavar="FILENAME",
        help="Box will envelop atoms in this file [.sdf .mol .mol2 .pdb .pdbqt]",
    )
    box_group.add_argument(
        "--padding", help="padding around atoms passed to --box_enveloping [A]", type=float
    )

    reactive_group = parser.add_argument_group("Reactive")
    reactive_group.add_argument(
        "-r",
        "--reactive_flexres",
        action="append",
        default=[],
        help='same as --flexres but for reactive residues (max 8)',
    )
    reactive_group.add_argument(
        "--reactive_name",
        action="append",
        default=[],
        help="set name of reactive atom of a residue type, e.g: --reactive_name 'TRP:NE1'. Repeat flag for multiple assignments. Overridden by --reactive_name_specific",
    )
    reactive_group.add_argument(
        "-s",
        "--reactive_name_specific",
        action="append",
        default=[],
        help="set name of reactive atom for an individual residue by the residue ID, e.g: -s 'A:42=NE2'. Residue will be reactive.",
    )

    reactive_group.add_argument(
        "--r_eq_12",
        default=1.8,
        type=float,
        help="r_eq for reactive atoms (1-2 interaction)",
    )
    reactive_group.add_argument(
        "--eps_12",
        default=2.5,
        type=float,
        help="epsilon for reactive atoms (1-2 interaction)",
    )
    reactive_group.add_argument(
        "--r_eq_13_scaling",
        default=0.5,
        type=float,
        help="r_eq scaling for 1-3 interaction across reactive atoms",
    )
    reactive_group.add_argument(
        "--r_eq_14_scaling",
        default=0.5,
        type=float,
        help="r_eq scaling for 1-4 interaction across reactive atoms",
    )
    args = parser.parse_args()
    
    if args.read_pdb is None and args.read_with_prody is None:
        parser.print_help()
        msg = "Need input filename: use either -i/--read_with_prody or --read_pdb"
        print(eol + msg)
        sys.exit(2)

    if args.read_pdb is not None and args.read_with_prody is not None:
        msg = "Can't use both -i/--read_with_prody and --read_pdb"
        print(eol + msg, file=sys.stderr)
        sys.exit(2)

    if args.write_gpf is not None and args.write_pdbqt is None:
        # there's a few of places that assume this condition has been checked
        msg = "--write_gpf requires --write_pdbqt because autogrid expects"
        msg += " the GPF file to point to the PDBQT file." 
        print(eol + msg)
        sys.exit(2)

    skip_gpf = args.write_gpf is None and args.write_vina_box is None
    if not skip_gpf:

        box_help = f"""
    writing a grid parameter file (--write_gpf) or a config file with the
    box dimensions for vina (-v/--write_vina_box) requires setting the box
    center and size with one of the following three combinations:
    1) --box_center and --box_size
    2) --box_center_off_reactive_res and --box_size
    3) --box_enveloping and --padding"""

        # Ensure correct number of box specs
        nr_boxcenter_specs = sum(
            [
                (args.box_center is not None),
                (args.box_center_off_reactive_res),
                (args.box_enveloping is not None),
            ]
        )
        nr_boxsize_specs = sum(
            [(args.box_size is not None), (args.padding is not None)]
        )

        box_specs = [(nr_boxcenter_specs, "box center"), (nr_boxsize_specs, "box size")]

        for spec_count, spec_type in box_specs:
            if spec_count > 1:
                msg = f"{spec_type} can't be specified in more than once. {box_help}"
                print("Command line error: " + msg, file=sys.stderr)
                sys.exit(2)
            elif spec_count < 1:
                msg = (
                    f"missing {spec_type} to write .gpf file for autogrid4. {box_help}"
                )
                print("Command line error: " + msg, file=sys.stderr)
                sys.exit(2)

        # Ensure correct combinations of box specs
        if args.box_size is None:
            if args.box_center_off_reactive_res:
                msg = f"--box_center_off_reactive_res requires --box_size. {box_help}"
                print("Command line error: " + msg, file=sys.stderr)
                sys.exit(2)
            elif args.box_center is not None:
                msg = f"--box_center requires --box_size. {box_help}"
                print("Command line error: " + msg, file=sys.stderr)
                sys.exit(2)

        if (args.padding is None) != (args.box_enveloping is None):
            msg = f"--padding and --box_enveloping must be used together. {box_help}"
            print("Command line error: " + msg, file=sys.stderr)
            sys.exit(2)

    return args


def main():
    args = get_args()
    
    if args.wanted_altloc is None:
        wanted_altloc = None
    else:
        wanted_altloc = parse_cmdline_res_assign(args.wanted_altloc)
        # Ensure meaningful wanted_altloc
        for key, value in wanted_altloc.items():
            if isinstance(value, str) and value.strip() == "":
                msg = "Wanted atloc cannot be an empty string or a string with just space"
                print("Command line error: " + msg, file=sys.stderr)
                sys.exit(2)
    
    
    # Ensure meaningful default_altloc
    if args.default_altloc is not None and args.default_altloc.strip()=="":
        msg = "Allowed atloc cannot be an empty string or a string with just space"
        print("Command line error: " + msg, file=sys.stderr)
        sys.exit(2)
    
    # check write options have default if used without argument
    write_flags = [
        args.write_pdbqt,
        args.write_json,
        args.write_gpf,
        args.write_vina_box,
    ]
    needed_default = False
    for flag in write_flags:
        # flag is none if not used, and is empty list when used without arg
        if flag is not None and len(flag) == 0:
            needed_default = True
            break
    if needed_default and args.output_basename is None:
        msg = "--write flags require either a filename argument or"
        msg += " --output_basename to set a default"
        print(msg)
        sys.exit(2)
    
    # Default mapping of residue name and reactive atom name
    reactive_atom = {
        "SER": "OG",
        "LYS": "NZ",
        "TYR": "OH",
        "CYS": "SG",
        "HIE": "NE2",
        "HID": "ND1",
        "GLU": "OE2",
        "THR": "OG1",
        "MET": "SD",
    }
    
    # Process custom mapping of residue name and reactive atom name
    modified = set()
    for react_name_str in args.reactive_name:
        resname, name = react_name_str.split(":")
        if resname in modified:
            print(
                "Command line error: repeated resname %s passed to --reactive_resname"
                % resname
                + eol,
                file=sys.stderr,
            )
            sys.exit(2)
        modified.add(resname)
        reactive_atom[resname] = name
    
    # Process specified mapping of residue ID and reactive atom name
    modified = set()
    reactive_flexres_name = {}
    for string in args.reactive_name_specific:
        res_assign = parse_cmdline_res_assign(string)
        for res_id in res_assign:
            if res_id in modified:
                print(
                    "Command line error: repeated resid %s passed to --reactive_name_specific"
                    % res_id
                    + eol,
                    file=sys.stderr,
                )
                sys.exit(2)
            modified.add(res_id)
            reactive_flexres_name[res_id] = res_assign[res_id]
    
    # Process residue ID of reactive flexible residues without specified reactive atom
    reactive_flexres = set(reactive_flexres_name)
    for resid_string in args.reactive_flexres:
        res_list = parse_cmdline_res(resid_string)
        for res_id in res_list:
            if res_id not in reactive_flexres:
                reactive_flexres.add(res_id)
                reactive_flexres_name[res_id] = ""
    
    # Evaluate number of reactive flexible residues
    if len(reactive_flexres) > 8:
        msg = "got %d reactive_flexres but maximum is 8." % (len(args.reactive_flexres))
        print("Command line error: " + msg, file=sys.stderr)
        sys.exit(2)
    
    # Evaluate compatibility with other options
    if len(reactive_flexres) != 1 and args.box_center_off_reactive_res:
        msg = (
            "--box_center_off_reactive_res can be used only with one reactive" + eol
        )
        msg += "residue, but %d reactive residues are set" % len(reactive_flexres_name)
        print("Command line error:" + msg, file=sys.stderr)
        sys.exit(2)
    
    # Process residue ID of nonreactive flexible residues
    nonreactive_flexres = set()
    for string in args.flexres:
        for res_id in parse_cmdline_res(string):
            if res_id not in reactive_flexres:
                nonreactive_flexres.add(res_id)
    
    
    set_template = {}
    if args.set_template is not None:
        set_template = parse_cmdline_res_assign(args.set_template)
    
    blunt_ends = []
    if args.blunt_ends is not None:
        j = parse_cmdline_res_assign(args.blunt_ends)
        # TODO parse also input/raw atom names, easier than indices
        j = [(k, int(v)) for k, v in j.items()]
        blunt_ends.extend(j)
    
    delete_residues = []
    if args.delete_residues is not None:
        delete_residues = parse_cmdline_res(args.delete_residues)
    
    if args.mk_config is not None:
        with open(args.mk_config) as f:
            mk_config = json.load(f)
        mk_prep = MoleculePreparation.from_config(mk_config)
    else:
        mk_prep = MoleculePreparation()
    
    # load templates for mapping
    templates = ResidueChemTemplates.create_from_defaults()
    for fn in args.add_templates:
        templates.add_json_file(fn)
    
    # create polymers
    if args.read_with_prody is not None:
        if not _got_prody:
            print(_prody_import_error, file=sys.stderr)
            print("option --read_with_prody requires Prody, which is not installed.")
            print("Installable from PyPI (pip install prody) or conda-forge (micromamba install prody)")
            sys.exit(2)
        ext = pathlib.Path(args.read_with_prody).suffix[1:].lower()
        if ext in SUPPORTED_PRODY_FORMATS:
            parser = SUPPORTED_PRODY_FORMATS[ext]
            input_obj = parser(args.read_with_prody, altloc="all")
            try:
                polymer = Polymer.from_prody(
                    input_obj,
                    templates,
                    mk_prep,
                    set_template,
                    delete_residues,
                    args.allow_bad_res,
                    blunt_ends=blunt_ends,
                    wanted_altloc=wanted_altloc,
                    default_altloc=args.default_altloc,
                )
            except PolymerCreationError as e:
                print(e)
                sys.exit(1)
    else:
        with open(args.read_pdb) as f:
            pdb_string = f.read()
        try:
            polymer = Polymer.from_pdb_string(
                pdb_string,
                templates,  # residue_templates, padders, ambiguous,
                mk_prep,
                set_template,
                delete_residues,
                args.allow_bad_res,
                blunt_ends=blunt_ends,
                wanted_altloc=wanted_altloc,
                default_altloc=args.default_altloc,
            )
        except PolymerCreationError as e:
            print(e)
            sys.exit(1)
    
    
    # Use residue name in the input structure file to find reactive atom name
    # According to the mapping of residue name and reactive atom name
    for res_id in reactive_flexres:
        if res_id not in polymer.monomers:
            print("resid %s not found in input receptor file" % res_id)
            sys.exit(2)
        res_atom = reactive_flexres_name[res_id]
        if not res_atom:
            input_resname = polymer.monomers[res_id].input_resname
            if input_resname in reactive_atom:
                reactive_flexres_name[res_id] = reactive_atom[input_resname]
            else:
                print("no default reactive name for %s, " % input_resname)
                print("use --reactive_name or --reactive_name_specific" + eol)
                sys.exit(2)
    
    # Print nonreactive and reactive flexible residues specs
    if len(nonreactive_flexres) + len(reactive_flexres) > 0:
        print()
        print("Flexible residues:")
        print("chain resnum is_reactive reactive_atom")
        string = "%5s%7s%12s%14s"
    
        if len(nonreactive_flexres) > 0:
            for res_id in nonreactive_flexres:
                chain, resnum = res_id.split(":")
                react_atom = ""
                print(string % (chain, resnum, False, react_atom))
    
        if len(reactive_flexres) > 0:
            for res_id in reactive_flexres_name:
                chain, resnum = res_id.split(":")
                react_atom = reactive_flexres_name[res_id]
                print(string % (chain, resnum, True, react_atom))
    
    # Assign reactive atom types for atoms in reactive flexible residues
    reactive_prefix = 1
    for res_id in reactive_flexres:
        # get reactive atom types
        reactive_aname = reactive_flexres_name[res_id]
        reactive_atomi = atom_name_to_molsetup_index(
            polymer.monomers[res_id], reactive_aname
        )
        if reactive_atomi is None:
            print(f"cannot find reactive atom name {reactive_aname} from residue {res_id} in input receptor file")
            sys.exit(2)
        reactive_atypes = assign_reactive_types_by_index(polymer.monomers[res_id].molsetup, reactive_atomi)
        # set reactive atom types
        nr_atom = len(polymer.monomers[res_id].molsetup.atoms)
        for atom_index in range(nr_atom):
            if (
                polymer.monomers[res_id].molsetup.atoms[atom_index].atom_type
                != reactive_atypes[atom_index]
            ):
                polymer.monomers[res_id].molsetup.atoms[
                    atom_index
                ].atom_type = f"{reactive_prefix}{reactive_atypes[atom_index]}"
        reactive_prefix += 1
    
    # Combine nonreactive and reactive flexible residues into one set
    all_flexres = nonreactive_flexres.union(reactive_flexres)
    
    for res_id in all_flexres:
        polymer.flexibilize_sidechain(res_id, mk_prep)
    
    
    any_lig_base_types = [
        "HD",
        "C",
        "A",
        "N",
        "NA",
        "OA",
        "F",
        "P",
        "SA",
        "S",
        "Cl",
        "CL",
        "Br",
        "BR",
        "I",
        "Si",
        "B",
    ]
    
    if args.output_basename is not None:
        outpath = pathlib.Path(args.output_basename)
    
    written_files_log = {"filename": [], "description": []}
    
    if args.write_json is not None:
        if args.write_json:
            fn = args.write_json[0]
        else:  # args.write_json is empty list (was used without arg)
            fn = str(outpath) + ".json"
        with open(fn, "w") as f:
            json.dump(polymer, f, cls=PolymerEncoder)
        written_files_log["filename"].append(fn)
        written_files_log["description"].append("parameterized receptor")
    
    if args.write_pdb is not None:
        fn = args.write_pdb[0]
        with open(fn, "w") as f:
            f.write(polymer.to_pdb())
        written_files_log["filename"].append(fn)
        written_files_log["description"].append("receptor")
    
    if args.write_pdbqt is not None:
        if args.write_pdbqt:
            if args.write_pdbqt[0].endswith(".pdbqt"):
                # may need to suffix _rigid/_flex with flexres
                fn_base = str(pathlib.Path(args.write_pdbqt[0]).with_suffix(""))
            else:
                fn_base = args.write_pdbqt[0]
        else:
            fn_base = str(outpath)
    
        pdbqt_tuple = PDBQTWriterLegacy.write_from_polymer(polymer)
        rigid_pdbqt, flex_pdbqt_dict = pdbqt_tuple
    
        if len(all_flexres) == 0:
            box_center = args.box_center
            rigid_fn = fn_base + ".pdbqt"
            flex_fn = None
        else:
            print(f"{reactive_flexres=}")
            all_flex_pdbqt = ""
            reactive_flexres_count = 0
            for res_id, flexres_pdbqt in flex_pdbqt_dict.items():
                all_flex_pdbqt += flexres_pdbqt
        
            rigid_fn = fn_base + "_rigid.pdbqt"
            flex_fn = fn_base + "_flex.pdbqt"
        
            written_files_log["filename"].append(flex_fn)
            written_files_log["description"].append("flexible receptor input file")
            with open(flex_fn, "w") as f:
                f.write(all_flex_pdbqt)
        
        written_files_log["filename"].append(rigid_fn)
        written_files_log["description"].append("static (i.e., rigid) receptor input file")
        with open(rigid_fn, "w") as f:
            f.write(rigid_pdbqt)
    
    def warn_flexres_outside_box(polymer, box_center, box_size):
        for res_id, res in polymer.monomers.items():
            if not res.is_movable:
                continue
            for atom in res.molsetup.atoms:
                if not res.is_flexres_atom[atom.index]:
                    continue
                if gridbox.is_point_outside_box(atom.coord, box_center, box_size, spacing=1.0):
                    print(
                        "WARNING: Flexible residue outside box." + eol,
                        file=sys.stderr,
                    )
                    print(
                        "WARNING: Strongly recommended to use a box that encompasses flexible residues."
                        + eol,
                        file=sys.stderr,
                    )
                    break  # only need to warn once
    
    skip_gpf = args.write_gpf is None and args.write_vina_box is None
    if not skip_gpf:
        if args.box_center is not None:
            box_center = args.box_center
            box_size = args.box_size
        elif args.box_center_off_reactive_res:
            # we have only one reactive residue and will set the box center
            # to be 5 Angstromg away from CB along the CA->CB vector
            box_centers = []
            for res_id in reactive_flexres:
                molsetup = polymer.monomers[res_id].molsetup
                calpha_idx = [
                    atom.index for atom in molsetup.atoms if atom.pdbinfo.name == "CA"
                ]
                cbeta_idx = [
                    atom.index for atom in molsetup.atoms if atom.pdbinfo.name == "CB"
                ]
                if len(calpha_idx) != 1:
                    check(
                        success=False,
                        error_msg=f"found {len(calpha_idx)} CA in {res_id} but expected 1",
                    )
                if len(cbeta_idx) != 1:
                    check(
                        success=False,
                        error_msg=f"found {len(cbeta_idx)} CB in {res_id} but expected 1",
                    )
                calpha_idx = calpha_idx[0]
                cbeta_idx = cbeta_idx[0]
                ca = molsetup.get_coord(calpha_idx)
                cb = molsetup.get_coord(cbeta_idx)
                v = cb - ca
                v /= math.sqrt(v[0] ** 2 + v[1] ** 2 + v[2] ** 2) + 1e-8
                box_center = ca + 5 * v
                box_centers.append(box_center)
            box_center = np.mean(box_centers, 0)
            box_size = args.box_size
        elif args.box_enveloping is not None:
            ft = pathlib.Path(args.box_enveloping).suffix
            suppliers = {
                ".pdb": Chem.MolFromPDBFile,
                ".mol": Chem.MolFromMolFile,
                ".mol2": Chem.MolFromMol2File,
                ".sdf": Chem.SDMolSupplier,
                ".pdbqt": None,
            }
            if ft not in suppliers.keys():
                check(
                    success=False,
                    error_msg=f"Given --box_enveloping file type {ft} not readable!"
                )
            elif ft != ".sdf" and ft != ".pdbqt":
                ligmol = suppliers[ft](args.box_enveloping, removeHs=False, sanitize=False)
            elif ft == ".sdf":
                ligmol = suppliers[ft](args.box_enveloping, removeHs=False, sanitize=False)[
                    0
                ]  # assume we only want first molecule in file
            else:  # .pdbqt
                ligmol = RDKitMolCreate.from_pdbqt_mol(
                    PDBQTMolecule.from_file(args.box_enveloping)
                )[
                    0
                ]  # assume we only want first molecule in file
    
            box_center, box_size = gridbox.calc_box(
                ligmol.GetConformer().GetPositions(), args.padding
            )
        else:
            print("Error: No box center specified.", file=sys.stderr)
            sys.exit(2)
    
    
        if args.write_gpf is not None:
            if args.write_gpf:
                gpf_fn = args.write_gpf[0]
            else:
                gpf_fn = pathlib.Path(rigid_fn).with_suffix(".gpf")
            # write .dat parameter file for B and Si
            ff_fn = pathlib.Path(gpf_fn).parents[0] / pathlib.Path(
                "boron-silicon-atom_par.dat"
            )
            written_files_log["filename"].append(str(ff_fn))
            written_files_log["description"].append(
                "atomic parameters for B and Si (for autogrid)"
            )
            with open(ff_fn, "w") as f:
                f.write(gridbox.boron_silicon_atompar)
    
            rec_types = [
                "HD",
                "C",
                "A",
                "N",
                "NA",
                "OA",
                "F",
                "P",
                "SA",
                "S",
                "Cl",
                "Br",
                "I",
                "Mg",
                "Ca",
                "Mn",
                "Fe",
                "Zn",
            ]
            gpf_string, npts = gridbox.get_gpf_string(
                box_center,
                box_size,
                pathlib.Path(rigid_fn).name,  # requires --write_pdbqt
                rec_types,
                any_lig_base_types,
                ff_param_fname=ff_fn.name,
            )
    
            written_files_log["filename"].append(str(gpf_fn))
            written_files_log["description"].append("autogrid input file")
            with open(gpf_fn, "w") as f:
                f.write(gpf_string)
    
        # write gridbox vina format
        if args.write_vina_box is not None:
            if args.write_vina_box:
                box_vina_fn = args.write_vina_box[0]
            else:
                box_vina_fn = str(outpath) + ".box.txt"
    
            written_files_log["filename"].append(box_vina_fn)
            written_files_log["description"].append("Vina-style box dimension file")
            with open(box_vina_fn, "w") as f:
                f.write(gridbox.box_to_vina_string(box_center, box_size))
    
        # write a PDB for the box
        if args.write_vina_box is not None or args.write_gpf is not None:
            if args.output_basename is not None:
                box_fn = str(outpath) + ".box.pdb" 
            elif args.write_gpf is not None:
                # relies on --write_gpf forcing --write_pdbqt which sets rigid_fn
                box_fn = str(pathlib.Path(rigid_fn).with_suffix(".box.pdb"))
            else:
                # suffix .pdb even if box_vina_fn does not end with ".txt"
                box_fn = box_vina_fn.replace(".txt", "") + ".pdb"
            written_files_log["filename"].append(box_fn)
            written_files_log["description"].append("PDB file to visualize the grid box")
            with open(box_fn, "w") as f:
                f.write(gridbox.box_to_pdb_string(box_center, box_size, spacing=1.0))
    
        warn_flexres_outside_box(polymer, box_center, box_size)
    
    
    # configuration info for AutoDock-GPU reactive docking
    if len(reactive_flexres) > 0 and args.write_pdbqt is not None:
        any_lig_reac_types = []
        for order in (1, 2, 3):
            for t in any_lig_base_types:
                any_lig_reac_types.append(reactive_typer.get_reactive_atype(t, order))
    
        rec_reac_types = []
        for line in all_flex_pdbqt.split(eol):
            if line.startswith("ATOM") or line.startswith("HETATM"):
                atype = line[77:].strip()
                basetype, _ = reactive_typer.get_basetype_and_order(atype)
                if basetype is not None:  # is None if not reactive
                    rec_reac_types.append(line[77:].strip())
    
        derivtypes, modpairs, collisions = get_reactive_config(
            any_lig_reac_types,
            rec_reac_types,
            args.eps_12,
            args.r_eq_12,
            args.r_eq_13_scaling,
            args.r_eq_14_scaling,
        )
    
        if len(collisions) > 0:
            collision_str = ""
            for t1, t2 in collisions:
                collision_str += "%3s %3s" % (t1, t2) + eol
            collision_fn = str(outpath.with_suffix(".atype_collisions"))
            written_files_log["filename"].append(collision_fn)
            written_files_log["description"].append(
                "type pairs (n=%d) that may lead to intra-molecular reactions"
                % len(collisions)
            )
            with open(collision_fn, "w") as f:
                f.write(collision_str)
    
        # The maps block is to tell AutoDock-GPU the base types for the reactive types.
        # This could be done with -T/--derivtypes, but putting derivtypes and intnbp
        # lines in a single configuration file simplifies the command line call.
        map_block = ""
        map_prefix = pathlib.Path(rigid_fn).with_suffix("").name
        all_types = []
        for basetype, reactypes in derivtypes.items():
            all_types.append(basetype)
            map_block += "map %s.%s.map" % (map_prefix, basetype) + eol
            for reactype in reactypes:
                all_types.append(reactype)
                map_block += "map %s.%s.map" % (map_prefix, basetype) + eol
        config = "ligand_types " + " ".join(all_types) + eol
        config += "fld %s.maps.fld" % map_prefix + eol
        config += map_block
    
        # in modpairs (dict): types are keys, parameters are values
        # now we will write a configuration file with nbp keywords
        # that AD-GPU reads using the --import_dpf flag
        # nbp stands for "non-bonded potential" or "non-bonded pairwise"
        line = "intnbp_r_eps %8.6f %8.6f %3d %3d %4s %4s" + eol
        nbp_count = 0
        for (t1, t2), param in modpairs.items():
            config += line % (param["r_eq"], param["eps"], param["n"], param["m"], t1, t2)
            nbp_count += 1
        config_fn = str(outpath.with_suffix(".reactive_config"))
        written_files_log["filename"].append(config_fn)
        written_files_log["description"].append("reactive parameters for AutoDock-GPU")
        with open(config_fn, "w") as f:
            f.write(config)
        print()
        print("For reactive docking, pass the configuration file to AutoDock-GPU:")
        print(
            "    autodock_gpu -C 1 --import_dpf %s --flexres %s -L <ligand_filename>"
            % (config_fn, flex_fn)
        )
        print()
    
    if written_files_log["filename"]:
        print()
        print("Files written:")
        longest_fn = max([len(fn) for fn in written_files_log["filename"]])
        line = "%%%ds <-- " % longest_fn + "%s"
        for fn, desc in zip(written_files_log["filename"], written_files_log["description"]):
            print(line % (fn, desc))
        if (
            args.output_basename is not None and
            args.output_basename.endswith(".pdbqt") and
            args.write_pdbqt is None
        ):
            print()
            print("PDBQT files were NOT written. Use -p/--write_pdbqt for that.")
            print("Note that -o/--output_basename just sets a default for --write flags")
            print()
    else:
        print()
        print()
        print("Receptor was prepared, but no files were written.")
        print("")
        print("Consider the following --write options:")
        print("  -p/--write_pdbqt")
        print("  -j/--write_json")
        print("  -g/--write_gpf")
        print("  -v/--write_vina_box")
        print("")
        print("Use -o/--output_basename, or set a filename after each --write flag")
        print("")
        print("Recommended for AutoDock-GPU:")
        print("  -o my_receptor -p -j -g")
        print("")
        print("Recommended for AutoDock-Vina:")
        print("  -o my_receptor -p -j -v")

if __name__ == "__main__":
    sys.exit(main())



from collections import namedtuple

# named tuple to contain information about an atom
PDBAtomInfo = namedtuple("PDBAtomInfo", "name resName resNum icode chain")
PDBResInfo = namedtuple("PDBResInfo", "resName resNum chain")  # used in obutils, maybe
